import os
import sys

# Highlight exception messages
# https://stackoverflow.com/questions/25109105/how-to-colorize-the-output-of-python-errors-in-the-gnome-terminal/52797444#52797444
try:
    import IPython.core.ultratb
except ImportError:
    # No IPython. Use default exception printing.
    pass
else:
    sys.excepthook = IPython.core.ultratb.ColorTB()

# Import submodules
from .math import *
from .plot_functions import *
from .config_validation import *

# Import subpackages
from .pysolver_view import *
from .properties import *
from .cycles import *
from .utilities import *

from .axial_turbine import compute_performance, compute_optimal_turbine
from . import centrifugal_compressor 

# Set plot options
set_plot_options()


# Package info
__version__ = "0.1.14"
PACKAGE_NAME = "turboflow"
URL_GITHUB = "https://github.com/turbo-sim/turboflow"
URL_DOCS = "https://turbo-sim.github.io/turboflow/"
URL_DTU = "https://thermalpower.dtu.dk/"
BREAKLINE = 80 * "-"


def print_banner():
    """Prints a banner."""
    banner = """
        ______ __  __ ____   ____   ____   ______ __    ____  _       __
       /_  __// / / // __ \ / __ ) / __ \ / ____// /   / __ \| |     / /
        / /  / / / // /_/ // __  |/ / / // /_   / /   / / / /| | /| / / 
       / /  / /_/ // _, _// /_/ // /_/ // __/  / /___/ /_/ / | |/ |/ /   
      /_/   \____//_/ |_|/_____/ \____//_/    /_____/\____/  |__/|__/  
    """
    print(BREAKLINE)
    print(banner)
    print(BREAKLINE)
    # https://manytools.org/hacker-tools/ascii-banner/
    # Style: Sland


def print_package_info():
    """Prints package information with predefined values."""

    info = f""" Version:       {__version__}
 Repository:    {URL_GITHUB}
 Documentation: {URL_DOCS}"""
    print_banner()
    print(BREAKLINE)
    print(info)
    print(BREAKLINE)


# def add_package_to_pythonpath():
#     """Append package containing this __file__ to the PYTHONPATH environment variable"""

#     # Define bash command to append cwd to PYTHONPATH
#     # PACKAGE_PATH = os.getcwd()
#     # PACKAGE_PATH = os.path.dirname(os.getcwd())
#     PACKAGE_PATH = os.path.dirname(__file__)
#     bashrc_header = f"# Append {PACKAGE_NAME} package to PYTHONPATH"
#     if sys.platform == "win32":  # Windows
#         bashrc_line = f'export PYTHONPATH=$PYTHONPATH\;"{PACKAGE_PATH}"'
#     else:  # Linux or MacOS
#         bashrc_line = f'export PYTHONPATH=$PYTHONPATH:"{PACKAGE_PATH}"'

#     # Locate the .bashrc file
#     bashrc_path = os.path.expanduser("~/.bashrc")

#     # Ask for user confirmation with default set to no
#     response = input(
#         f"Do you want to add the {PACKAGE_NAME} path to your .bashrc? [yes/NO]: "
#     )
#     if response.lower() in ["y", "yes"]:
#         try:
#             # Check if the line already exists in the .bashrc
#             with open(bashrc_path, "r") as file:
#                 if bashrc_line in file.read():
#                     print(f".bashrc already contains the {PACKAGE_NAME} path.")
#                 else:
#                     with open(bashrc_path, "a") as file_append:
#                         file_append.write(f"\n{bashrc_header}")
#                         file_append.write(f"\n{bashrc_line}\n")
#                     print(f"Path to {PACKAGE_NAME} package added to .bashrc")
#                     print(
#                         f"Restart the terminal or run 'source ~/.bashrc' for the changes to take effect."
#                     )
#         except Exception as e:
#             print(f"An error occurred: {e}")
#     else:
#         print("Operation aborted by user.")




import numpy as np
from .. import math

def compute_losses(input_parameters, component):

    # Map loss model functions
    component_function = {
        "impeller" : compute_impeller_losses,
        "vaneless_diffuser" : compute_vaneless_diffuser_losses,
    }   

    loss_dict = component_function[component](input_parameters)

    return loss_dict


def compute_vaneless_diffuser_losses(input_parameters):

    """
    Stanitz vanless diffuser losses
    """
    
    # Load parameters
    T0_in = input_parameters["inlet_plane"]["T0"]
    p0_in = input_parameters["inlet_plane"]["p0"]
    v_in = input_parameters["inlet_plane"]["v"]
    cp = input_parameters["inlet_plane"]["cp"]
    cv = input_parameters["inlet_plane"]["cv"]
    p_out = input_parameters["exit_plane"]["p"]
    p0_out = input_parameters["exit_plane"]["p0"]

    # Calculate vaneless diffuser losses (Stanitz)
    gamma = cp/cv
    alpha = (gamma - 1)/gamma
    Y_tot = cp*T0_in*((p_out/p0_out)**alpha - (p_out/p0_in)**alpha)

    # Store losses in dict
    scale = 0.5*v_in**2
    loss_dict = {
        "loss_total" : Y_tot/scale
    }

    return loss_dict

def compute_impeller_losses(input_parameters):

    """
    Zhang set 1 Impeller losses
    Throat not evaluated for now
    """
    
    # Load inlet plane
    v_in = input_parameters["inlet_plane"]["v"]
    v_m_in = input_parameters["inlet_plane"]["v_m"]
    w_in = input_parameters["inlet_plane"]["w"]
    w_t_in = input_parameters["inlet_plane"]["w_t"]
    beta_in = input_parameters["inlet_plane"]["beta"]
    d_in = input_parameters["inlet_plane"]["d"]
    h0_in = input_parameters["inlet_plane"]["h0"]
    a_in = input_parameters["inlet_plane"]["a"]
    p_in = input_parameters["inlet_plane"]["p"]
    Ma_rel_in = input_parameters["inlet_plane"]["Ma_rel"]
    mu_in = input_parameters["inlet_plane"]["mu"]
    p0_in = input_parameters["inlet_plane"]["p0"]
    T0_in = input_parameters["inlet_plane"]["T0"]
    gamma_in = input_parameters["inlet_plane"]["gamma"]
    cp_in = input_parameters["inlet_plane"]["cpmass"]
    cv_in = input_parameters["inlet_plane"]["cvmass"]

    # Load exit plane
    v_out = input_parameters["exit_plane"]["v"]
    v_m_out = input_parameters["exit_plane"]["v_m"]
    v_t_out = input_parameters["exit_plane"]["v_t"]
    alpha_out = input_parameters["exit_plane"]["alpha"]
    beta_out = input_parameters["exit_plane"]["beta"]
    w_out = input_parameters["exit_plane"]["w"]
    w_t_out = input_parameters["exit_plane"]["w_t"]
    u_out = input_parameters["exit_plane"]["blade_speed"]
    d_out = input_parameters["exit_plane"]["d"]
    h0_out = input_parameters["exit_plane"]["h0"]
    a_out = input_parameters["exit_plane"]["a"]

    # Load throat plane
    w_th = input_parameters["throat_plane"]["w"]
    p_th = input_parameters["throat_plane"]["p"]

    # Load geometry
    t_cl = input_parameters["geometry"]["tip_clearance"]
    b_out = input_parameters["geometry"]["width_out"]
    L_ax = input_parameters["geometry"]["length_axial"]
    z = input_parameters["geometry"]["number_of_blades"]
    r_in_hub = input_parameters["geometry"]["radius_hub_in"]
    r_in_tip = input_parameters["geometry"]["radius_tip_in"]
    r_out = input_parameters["geometry"]["radius_out"]
    A_out = input_parameters["geometry"]["area_out"]
    A_in = input_parameters["geometry"]["area_in"]
    A_th = input_parameters["geometry"]["area_throat"]
    theta_in = input_parameters["geometry"]["leading_edge_angle"]

    # Load boundary conditions
    mass_flow_rate = input_parameters["boundary_conditions"]["mass_flow_rate"]

    # Load model factors
    wake_width = input_parameters["factors"]["wake_width"]

    # Incidence loss (Aungier)
    Y_inc = 0.4*(w_in-v_m_in/math.cosd(theta_in))**2 

    # Blade loading loss (Aungier)
    L_b = np.pi/8*(r_out*2-(r_in_tip + r_in_hub)-b_out+2*L_ax)*(2/(math.cosd(beta_in)+math.cosd(beta_out)))
    delta_W = 2*np.pi*2*r_out*v_t_out/(z*L_b)
    Y_bld = delta_W**2/48

    # Skin friction loss (Jansen)
    D_h = 2*r_out*math.cosd(beta_out)/(z/np.pi + 2*r_out*math.cosd(beta_out)/b_out) + 0.5*(r_in_tip+r_in_hub)/r_out*math.cosd(beta_in)/(z/np.pi + (r_in_tip + r_in_hub)/(r_in_tip-r_in_hub)*math.cosd(beta_in))
    w_avg = (2*w_in + 2*w_in)/4
    Re = u_out*D_h*d_in/mu_in
    Cf = 0.0412*Re**(-0.1925)
    Y_sf = 2*Cf*L_b/D_h*w_avg**2

    # Clearance loss (Rogers)
    Y_cl = 0.1*t_cl/b_out*u_out**2

    # Mixing loss (Aungier)
    W_out = np.sqrt((v_m_out*A_out/(2*np.pi*r_out*b_out))**2+w_t_out**2)
    W_max = (w_in + w_out+delta_W)/2
    D_eq = W_max/w_out
    W_sep = w_out*(D_eq <= 2.0) + w_out*D_eq/2*(D_eq > 2.0) 
    Y_mix = 0.5*(W_sep - W_out)**2

    # Entrance diffusion loss (Aungier)
    Y_dif = max(0, 0.4*(w_in - w_th)**2 - Y_inc)

    # Choke loss (Aungier)
    R = cp_in-cv_in
    A_crit = mass_flow_rate/(p0_in*np.sqrt(gamma_in/(R*T0_in))*(2/(gamma_in+1))**((gamma_in+1)/(2*(gamma_in-1))))
    Cr = np.sqrt(A_in*math.cosd(theta_in)/A_th)
    Cr = min(Cr, 1-(A_in*math.cosd(theta_in)/A_th-1)**2)
    X = 11-10*(Cr*A_th)/A_crit
    Y_ch = 0.5*w_in**2*(0.05*X+X**7)*(X > 0)

    # Shock losses (Whitfield and Baines)
    Y_sh = w_in**2/2*(1-(w_th/w_in)**2-2/((gamma_in-1)*Ma_rel_in**2)*((p_th/p_in)**(gamma_in/(gamma_in-1))-1))

    # Parasitic losses
    parasitic_losses = compute_parasitic_losses(input_parameters)
    Y_lk = parasitic_losses["leakage"]
    Y_df = parasitic_losses["disk_friction"]
    Y_rc = parasitic_losses["recirculation"]

    # Total losses
    scale = 0.5*v_in**2
    Y_inc = Y_inc/scale
    Y_bld = Y_bld/scale
    Y_sf = Y_sf/scale
    Y_cl = Y_cl/scale
    Y_mix = Y_mix/scale

    Y_dif = Y_dif/scale
    Y_ch = Y_ch/scale
    Y_sh = Y_sh/scale

    Y_lk = Y_lk/scale
    Y_df = Y_df/scale
    Y_rc = Y_rc/scale

    Y_parasitic = Y_lk + Y_df + Y_rc
    Y_tot = Y_inc + Y_bld + Y_sf + Y_cl + Y_mix + Y_dif + Y_ch + Y_sh + Y_parasitic

    # Scale loss coefficent
    loss_dict = {"incidence" : Y_inc,
                 "blade_loading" : Y_bld,
                 "skin_friction" : Y_sf,
                 "tip_clearance" : Y_cl,
                 "wake_mixing" : Y_mix,
                 "loss_total" : Y_tot,
                 "leakage" : Y_lk,
                 "recirculation" : Y_rc,
                 "disk_friction" : Y_df,
                 "choke": Y_ch,
                 "shock" : Y_sh,
                 "entrance_diffusion" : Y_dif}
    
    return loss_dict

def compute_parasitic_losses(input_parameters):

    """
    Zhang set 1 parasitic losses
    """

    # Load boundary conditions 
    mass_flow_rate = input_parameters["boundary_conditions"]["mass_flow_rate"]

    # Load inlet plane
    v_in = input_parameters["inlet_plane"]["v"]
    v_t_in = input_parameters["inlet_plane"]["v_t"]
    u_in = input_parameters["inlet_plane"]["blade_speed"]
    w_in = input_parameters["inlet_plane"]["w"]
    d_in = input_parameters["inlet_plane"]["d"]
    h0_in = input_parameters["inlet_plane"]["h0"]

    # Load exit plane
    v_t_out = input_parameters["exit_plane"]["v_t"]
    alpha_out = input_parameters["exit_plane"]["alpha"]
    w_out = input_parameters["exit_plane"]["w"]
    v_out = input_parameters["exit_plane"]["v"]
    u_out = input_parameters["exit_plane"]["blade_speed"]
    d_out = input_parameters["exit_plane"]["d"]
    h0_out = input_parameters["exit_plane"]["h0"]
    mu_out = input_parameters["exit_plane"]["mu"]

    # Load geometry
    t_cl = input_parameters["geometry"]["tip_clearance"]
    b_out = input_parameters["geometry"]["width_out"]
    b_in = input_parameters["geometry"]["width_in"]
    L_ax = input_parameters["geometry"]["length_axial"]
    L_m = input_parameters["geometry"]["length_meridional"]
    z = input_parameters["geometry"]["number_of_blades"]
    r_in_tip = input_parameters["geometry"]["radius_tip_in"]
    r_hub_in = input_parameters["geometry"]["radius_hub_in"]
    r_out = input_parameters["geometry"]["radius_out"]
    r_mean_in = input_parameters["geometry"]["radius_mean_in"]

    # Recirculation losses (Coppage)
    Df = 1 - w_out/w_in + 0.75*(abs(u_out*v_t_out) - abs(u_in*v_t_in))*w_out/((z/np.pi*(1-r_in_tip/r_out)+2*r_in_tip/r_out)*w_in*u_out**2)
    Y_rc = 0.02*np.sqrt(math.tand(alpha_out))*Df**2*u_out**2 # Error in exponent in paper?

    # Disk friction losses (Dailey and Nece)
    Re = d_out*u_out*r_out/mu_out
    f_df = 2.67/Re**0.5*(Re<3e5) + 0.0622/Re**0.2*(Re>=3e5)
    d_avg = (d_in + d_out)/2
    Y_df = f_df*d_avg*r_out**2*u_out**3/(4*mass_flow_rate)

    # Leakage loss (Jansen)
    alpha = (r_in_tip-r_hub_in)/(r_out-r_in_tip)/(1+d_out/d_in)
    Y_lk = 0.6*t_cl/b_out*v_out*np.sqrt(4*np.pi/b_out/z*alpha*v_t_out*v_in)

    loss_dict = {"recirculation" : Y_rc,
                 "disk_friction" : Y_df,
                 "leakage" : Y_lk,
                 "loss_total" : Y_rc + Y_df + Y_lk,
                 }
    
    return loss_dict



import numpy as np
import pandas as pd
import CoolProp as cp
from . import loss_model as lm
from . import slip_model as sm
from . import choking_criterion as cm
from .. import utilities as utils
from .. import math
import scipy.linalg
import scipy.integrate
from scipy.optimize._numdiff import approx_derivative

def evaluate_centrifugal_compressor(
    variables,
    boundary_conditions,
    geometry,
    fluid,
    model_options,
    reference_values,
):

    # Initialize results structure
    compressor_results = {"residuals" : {},
                          "planes" : pd.DataFrame(),
                          "boundary_conditions" : boundary_conditions}

    # Map functions for each component
    component_function = {
        "impeller" : evaluate_impeller,
        "vaneless_diffuser" : evaluate_vaneless_diffuser,
        "vaned_diffuser" : evaluate_vaned_diffuser,
        "volute" : evaluate_volute,
    }

    # Prepare first set of inputs
    input = {"h0_in" : boundary_conditions["h0_in"],
             "s_in" : boundary_conditions["s_in"],
             "alpha_in" : boundary_conditions["alpha_in"],
             }

    # Evaluate components
    for key in geometry.keys():
        # Prepare set of independent variables
        independents = {var : val for var, val in variables.items() if var.endswith(key)}

        # Evaluate component
        results, residuals = component_function[key](independents, input, boundary_conditions, geometry[key], fluid, model_options, reference_values)

        # Store result in result structure
        compressor_results[key] = results
        compressor_results["planes"].loc[len(compressor_results["planes"]), results["inlet_plane"].keys()] = results["inlet_plane"]
        compressor_results["planes"].loc[len(compressor_results["planes"]), results["exit_plane"].keys()] = results["exit_plane"]

        residuals = utils.add_string_to_keys(residuals, f"_{key}")
        compressor_results["residuals"].update(residuals)

        # Prepare calculation of the subsequent component
        input = compressor_results["planes"].loc[len(compressor_results["planes"])-1].to_dict()

    # Compute overall performance
    compressor_results["overall"] = compute_overall_performance(compressor_results, fluid, geometry)

    return compressor_results

def evaluate_impeller(independents, input, boundary_conditions, geometry, fluid, model_options, reference_values):
    """
    Evaluate impeller performance
    - Deviation calculated according to slip factor
    - Losses calculated according to correlation for loss coefficient
    """

    # Load reference_values
    v_max = reference_values["v_max"]
    s_range = reference_values["s_range"]
    s_min = reference_values["s_min"]
    angle_range = reference_values["angle_range"]
    angle_min = reference_values["angle_min"]

    # Load independent variables
    v_in = independents["v_in_impeller"]*v_max
    w_out = independents["w_out_impeller"]*v_max 
    beta_out = independents["beta_out_impeller"]*angle_range + angle_min 
    s_out = independents["s_out_impeller"]*s_range + s_min

    # Load input variables
    h0_in = input["h0_in"]
    s_in = input["s_in"]
    alpha_in = input["alpha_in"]

    # Load boundary conditions
    mass_flow_rate = boundary_conditions["mass_flow_rate"]
    omega = boundary_conditions["omega"]

    # Load impeller geometry
    A_in = geometry["area_in"]
    A_out = geometry["area_out"]
    r_mean_in = geometry["radius_mean_in"]
    r_out = geometry["radius_out"]
    theta_out = geometry["trailing_edge_angle"]

    # Load model options
    slip_model = model_options ["slip_model"]
    loss_model = model_options ["loss_model"]["impeller"]

    # Evaluate inlet velocity triangle
    u_in = omega*r_mean_in
    v_m_in = v_in*math.cosd(alpha_in)
    v_t_in = v_in*math.sind(alpha_in)
    w_m_in = v_m_in
    w_t_in = v_t_in - u_in
    w_in = np.sqrt(w_t_in**2 + w_m_in**2)
    beta_in = math.arctand(w_t_in/w_m_in)

    velocity_triangle_in = {"v_t" : v_t_in,
                   "v_m" : v_m_in,
                   "v" : v_in,
                   "alpha" : alpha_in,
                   "w_t" : w_t_in,
                   "w_m" : w_m_in,
                   "w" : w_in,
                   "beta" : beta_in,
                   "blade_speed" : u_in,
                   }

    # Evaluate inlet thermodynamic state
    h_in = h0_in - 0.5*v_in**2
    h0_rel_in = h_in + 0.5*w_in**2
    static_properties_in = fluid.get_props(cp.HmassSmass_INPUTS, h_in, s_in)
    stagnation_properties_in = fluid.get_props(cp.HmassSmass_INPUTS, h0_in, s_in)
    relative_properties_in = fluid.get_props(cp.HmassSmass_INPUTS, h0_rel_in, s_in)

    # Evaluate inlet mass flow rate, rothalpy and mach
    m_in = w_m_in*static_properties_in["d"]*A_in
    rothalpy = h0_rel_in - 0.5*u_in**2
    Ma_in = v_in/static_properties_in["a"]
    Ma_rel_in = w_in/static_properties_in["a"]

    # Evaluate exit velocity triangle
    u_out = omega*r_out
    w_m_out = w_out*math.cosd(beta_out)
    w_t_out = w_out*math.sind(beta_out)
    v_m_out = w_m_out
    v_t_out = w_t_out + u_out
    v_out = np.sqrt(v_t_out**2 + v_m_out**2)
    alpha_out = math.arctand(v_t_out/v_m_out)

    velocity_triangle_out = {"v_t" : v_t_out,
                   "v_m" : v_m_out,
                   "v" : v_out,
                   "alpha" : alpha_out,
                   "w_t" : w_t_out,
                   "w_m" : w_m_out,
                   "w" : w_out,
                   "beta" : beta_out,
                   "blade_speed" : u_out,
                   }

    # Evaluate inlet thermodynamic state
    h0_rel_out = rothalpy + 0.5*u_out**2
    h_out = h0_rel_out - 0.5*w_out**2
    h0_out = h_out + 0.5*v_out**2
    static_properties_out = fluid.get_props(cp.HmassSmass_INPUTS, h_out, s_out)
    stagnation_properties_out = fluid.get_props(cp.HmassSmass_INPUTS, h0_out, s_out)
    relative_properties_out = fluid.get_props(cp.HmassSmass_INPUTS, h0_rel_out, s_out)  
    isentropic_relative_properties_out = fluid.get_props(cp.PSmass_INPUTS, relative_properties_out["p"], s_in)
    isentropic_static_properties_out = fluid.get_props(cp.PSmass_INPUTS, static_properties_out["p"], s_in)
    isentropic_properties_out = {**isentropic_static_properties_out,**utils.add_string_to_keys(isentropic_relative_properties_out, "0")}

    # Evaluate exit mass flow rate and Mach 
    m_out = static_properties_out["d"]*w_m_out*A_out
    Ma_out = v_out/static_properties_out["a"]
    Ma_rel_out = w_out/static_properties_out["a"]

    # Evaluate throat residual
    input_choking = {
        "s_in" : s_in,
        "rothalpy" : rothalpy,
    }
    throat_residual, throat_plane = cm.evaluate_choking(independents, input_choking, boundary_conditions, geometry, fluid, model_options, reference_values, component="impeller", u_throat = u_in)

    # Collect inlet plane and exit plane results 
    inlet_plane = {**velocity_triangle_in,
                   **static_properties_in,
                   **utils.add_string_to_keys(stagnation_properties_in, "0"),
                   **utils.add_string_to_keys(relative_properties_in, "0_rel"),
                   "Ma" : Ma_in,
                   "Ma_rel" : Ma_rel_in,
                   }
    
    exit_plane = {**velocity_triangle_out,
                   **static_properties_out,
                   **utils.add_string_to_keys(stagnation_properties_out, "0"),
                   **utils.add_string_to_keys(relative_properties_out, "0_rel"),
                    "Ma" : Ma_out,
                   "Ma_rel" : Ma_rel_out,
                   }

    # Evaluate loss coefficient
    loss_input = {
        "inlet_plane" : inlet_plane,
        "exit_plane" : exit_plane,
        "throat_plane" : throat_plane,
        "geometry" : geometry,
        "isentropic" : isentropic_properties_out,
        "factors" : model_options["factors"],
        "boundary_conditions" : boundary_conditions,
    }
    loss_dict = lm.evaluate_loss_model(loss_model, loss_input, "impeller")
    exit_plane = {**exit_plane, **loss_dict}

    # Evaluate slip
    slip_velocity = u_out + v_m_out*math.tand(theta_out) - v_t_out
    slip_input = {"u_out" : u_out}
    slip_model = sm.evaluate_slip(slip_model, slip_input, geometry)

    # Evaluate residuals
    resiudals = {"mass_flow_in" : (m_in-mass_flow_rate)/mass_flow_rate,
                 "mass_flow_out" : (m_out-mass_flow_rate)/mass_flow_rate,
                 "losses" : loss_dict["loss_error"],
                 "slip" : (slip_velocity-slip_model)/v_max,
                 **throat_residual}

    # Store impeller results
    results = {"inlet_plane" : inlet_plane,
               "throat_plane" : throat_plane,
               "exit_plane": exit_plane,
    }

    return results, resiudals

def evaluate_vaned_diffuser(independents, input, boundary_conditions, geometry, fluid, model_options, reference_values):

    """
    Evaluate vaned diffuser
    - Deviation calculated according to aungier 
    - Losses calculated according to correlation for loss coefficient
    """

    # Load reference_values
    v_max = reference_values["v_max"]
    s_range = reference_values["s_range"]
    s_min = reference_values["s_min"]

    # Load independent variables
    v_out = independents["v_out_vaned_diffuser"]*v_max
    s_out = independents["s_out_vaned_diffuser"]*s_range + s_min
    
    # Load input variables
    alpha_in = input["alpha"]
    h0_in = input["h0"]
    s_in = input["s"]

    # Load diffuser geometry
    theta_in = geometry["leading_edge_angle"]
    theta_out = geometry["trailing_edge_angle"]
    loc_camber_max = geometry["loc_camber_max"]
    camber = geometry["camber_angle"]
    solidity = geometry["solidity"]
    A_out = geometry["area_out"]

    # Load boundary conditions
    mass_flow_rate = boundary_conditions["mass_flow_rate"]

    # Load model options
    loss_model = model_options["loss_model"]["vaned_diffuser"]

    # Calculate incidence
    alpha_inc = alpha_in - theta_in

    # Calculate deviation angle and absolute flow angle
    # delta_0 = camber*(0.92*loc_camber_max**2 + 0.02*theta_out)/(np.sqrt(solidity) - 0.02*camber)
    # d_delta = np.exp(((1.5-(90-theta_in)/60)**2-3.3)*solidity)
    # alpha_out = theta_out + delta_0 + d_delta*alpha_inc
    # dev = {"delta_0" : delta_0,
    #     "alpha_inc" : alpha_inc,
        # "d_delta" : d_delta,
    #     }
    alpha_out = theta_out

    # Calculate tangential and meridional velocity
    v_m_out = v_out*math.cosd(alpha_out)
    v_t_out = v_out*math.sind(alpha_out)

    velocity_triangle_out = {"v_t" : v_t_out,
                   "v_m" : v_m_out,
                   "v" : v_out,
                   "alpha" : alpha_out,
                   }

    # Calculate thermophysical properties
    h0_out = h0_in
    h_out = h0_out - 0.5*v_out**2
    static_properties_out = fluid.get_props(cp.HmassSmass_INPUTS, h_out, s_out)
    stagnation_properties_out = fluid.get_props(cp.HmassSmass_INPUTS, h0_out, s_out)
    isentropic_stagnation_properties_out = fluid.get_props(cp.PSmass_INPUTS, stagnation_properties_out["p"], s_in)
    isentropic_static_properties_out = fluid.get_props(cp.PSmass_INPUTS, static_properties_out["p"], s_in)
    isentropic_properties_out = {**isentropic_static_properties_out,**utils.add_string_to_keys(isentropic_stagnation_properties_out, "0")}

    # Calculate mass flow rate and mach
    m_out = static_properties_out["d"]*v_m_out*A_out
    Ma_out = v_out/static_properties_out["a"]

    # Evaluate throat residual
    input_choking = {
        "s_in" : s_in,
        "rothalpy" : h0_in,
    }
    throat_residual, throat_results = cm.evaluate_choking(independents, input_choking, boundary_conditions, geometry, fluid, model_options, reference_values, component="vaned_diffuser")


    # Collect exit plane results
    exit_plane = {**velocity_triangle_out,
                   **static_properties_out,
                   **utils.add_string_to_keys(stagnation_properties_out, "0"),
                   "Ma" : Ma_out,
                   }

    # Evaluate loss coefficient
    loss_input = {
        "inlet_plane" : input,
        "exit_plane" : exit_plane,
        "geometry" : geometry,
        "isentropic" : isentropic_properties_out,
        "factors" : model_options["factors"],
        "boundary_conditions" : boundary_conditions,
    }
    loss_dict = lm.evaluate_loss_model(loss_model, loss_input, "vaned_diffuser")
    exit_plane = {**exit_plane, **loss_dict}

    # Evaluate residuals
    resiudals = {"mass_flow_out" : (m_out-mass_flow_rate)/mass_flow_rate,
                 "losses" : loss_dict["loss_error"],
                 **throat_residual
    }

    # Store diffuser results
    results = {"inlet_plane" : input,
               "throat_plane" : throat_results,
               "exit_plane": exit_plane,
               }
    
    return results, resiudals


def evaluate_vaneless_diffuser(independents, input, boundary_conditions, geometry, fluid, model_options, reference_values):

    vaneless_diffuser_model = model_options["vaneless_diffuser_model"]

    diffuser_function = {
        "algebraic" : evaluate_vaneless_diffuser_algebraic,
        "1D" : evaluate_vaneless_diffuser_1D,
    }

    if vaneless_diffuser_model in diffuser_function.keys():
        results, residual = diffuser_function[vaneless_diffuser_model](independents, input, boundary_conditions, geometry, fluid, model_options, reference_values)
        return results, residual
    else:
        options = ", ".join(f"'{k}'" for k in diffuser_function.keys())
        raise ValueError(
            f"Invalid diffuser model: '{vaneless_diffuser_model}'. Available options: {options}"
        )

def evaluate_vaneless_diffuser_algebraic(independents, input, boundary_conditions, geometry, fluid, model_options, reference_values):

    """
    Evaluate vaneless diffuser using algerbraic equations
    - Use conservation of angular momentum, accouting for friction by introducing a loss term
    - Pressure losses calculated according to correlation for loss coefficient
    """

    # Load reference_values
    v_max = reference_values["v_max"]
    s_range = reference_values["s_range"]
    s_min = reference_values["s_min"]
    angle_range = reference_values["angle_range"]
    angle_min = reference_values["angle_min"]

    # Load independent variables
    v_out = independents["v_out_vaneless_diffuser"]*v_max
    s_out = independents["s_out_vaneless_diffuser"]*s_range + s_min 
    alpha_out = independents["alpha_out_vaneless_diffuser"]*angle_range + angle_min

    # Load input variables
    h0_in = input["h0"]
    s_in = input["s"]
    v_in = input["v"]
    v_t_in = input["v_t"]
    alpha_in = input["alpha"]

    # Load diffuser geometry
    b_in = geometry["width_in"]
    r_in = geometry["radius_in"]
    r_out = geometry["radius_out"]
    A_out = geometry["area_out"]

    # Load boundary conditions
    mass_flow_rate = boundary_conditions["mass_flow_rate"]

    # Load model factors
    Cf = model_options["factors"]["skin_friction"]
    loss_model = model_options["loss_model"]["vaneless_diffuser"]

    # Calculate tangential and meridional velocity
    v_m_out = v_out*math.cosd(alpha_out)
    v_t_out = v_out*math.sind(alpha_out)

    velocity_triangle_out = {"v_t" : v_t_out,
                   "v_m" : v_m_out,
                   "v" : v_out,
                   "alpha" : alpha_out,
                   }
    
    # Calculate thermophysical properties
    h0_out = h0_in
    h_out = h0_out - 0.5*v_out**2
    static_properties_out = fluid.get_props(cp.HmassSmass_INPUTS, h_out, s_out)
    stagnation_properties_out = fluid.get_props(cp.HmassSmass_INPUTS, h0_out, s_out)
    isentropic_stagnation_properties_out = fluid.get_props(cp.PSmass_INPUTS, stagnation_properties_out["p"], s_in)
    isentropic_static_properties_out = fluid.get_props(cp.PSmass_INPUTS, static_properties_out["p"], s_in)
    isentropic_properties_out = {**isentropic_static_properties_out,**utils.add_string_to_keys(isentropic_stagnation_properties_out, "0")}

    # Calculate mass flow rate and mach 
    m_out = static_properties_out["d"]*v_m_out*A_out
    Ma_out = v_out/static_properties_out["a"]

    # Collect exit plane results
    exit_plane = {**velocity_triangle_out,
                   **static_properties_out,
                   **utils.add_string_to_keys(stagnation_properties_out, "0"),
                   "Ma" : Ma_out
                   }

    # Evaluate loss coefficient
    loss_input = {
        "inlet_plane" : input,
        "exit_plane" : exit_plane,
        "geometry" : geometry,
        "isentropic" : isentropic_properties_out,
        "factors" : model_options["factors"],
        "boundary_conditions" : boundary_conditions,
    }
    loss_dict = lm.evaluate_loss_model(loss_model, loss_input, "vaneless_diffuser")
    exit_plane = {**exit_plane, **loss_dict}

    # Evaluate change of angular momentum
    phi_is = r_out*v_t_out/(r_in*v_t_in)
    phi_AM = np.exp(-Cf*(r_out-r_in)/(b_in*math.cosd(alpha_in)))

    # Evaluate residuals
    resiudals = {"mass_flow_out" : (m_out-mass_flow_rate)/mass_flow_rate,
                 "losses" : loss_dict["loss_error"],
                 "angular_momentum" : phi_is - phi_AM,
    }

    # Store diffuser results
    results = {"inlet_plane" : input,
               "exit_plane": exit_plane}
    
    return results, resiudals

def evaluate_vaneless_diffuser_1D(independents, input, boundary_conditions, geometry, fluid, model_options, reference_values):
    
    """
    Evaluate vaneless diffuser using 1D transport equations
    """

    # Rename inlet state properties
    v_m_in = input["v_m"]
    v_t_in = input["v_t"]
    d_in = input["d"]
    p_in = input["p"]

    # Rename geometry
    div = geometry["wall_divergence"]
    r_in = geometry["radius_in"]
    r_out = geometry["radius_out"]
    b_in = geometry["width_in"]
    phi = 0 # Diffuser channel is radial

    # Load model options
    Cf = model_options["factors"]["skin_friction"]
    q_w = model_options["factors"]["wall_heat_flux"]

    # Define integration interval
    m_out = r_out - r_in
    
    def odefun(t, y):

        # Rename from ODE terminology to physical variables
        m = t     
        v_m, v_t, d, p = y

        # Calculate velocity
        v = np.sqrt(v_t**2 + v_m**2)
        alpha = math.arctand(v_t/v_m)

        velocity_triangle = {"v_t" : v_t,
                             "v_m" : v_m,
                             "v" : v,
                             "alpha" : alpha}

        # Calculate local radius
        r = r_fun(r_in, phi, m)
        b = b_fun(b_in, div, m)

        # Calculate derivative of area change
        delta = 1e-4
        diff_br = (b_fun(b_in,div,m+delta)*r_fun(r_in,phi,m+delta) - b*r)/delta

        # Calculate derivative of internal energy
        delta = 1e-4
        e1 = fluid.get_props(cp.DmassP_INPUTS, d, p-delta)
        e1 = e1["u"]
        e2 = fluid.get_props(cp.DmassP_INPUTS, d, p+delta)
        e2 = e2["u"]
        dedp_d = (e2 - e1)/(2*delta)

        # Calculate thermodynamic state
        static_properties = fluid.get_props(cp.DmassP_INPUTS, d, p)
        a = static_properties["a"]
        h = static_properties["h"]
        s = static_properties["s"]
        h0 = h + 0.5*v**2
        stagnation_properties = fluid.get_props(cp.HmassSmass_INPUTS, h0, s)

        # Stress at the wall
        tau_w = Cf*d*v**2/2

        # Compute coefficient matrix
        M = np.asarray(
            [
                [d, 0.0, v_m, 0],
                [d*v_m, 0.0, 0.0, 1.0],
                [0.0, d*v_m, 0.0, 0.0],
                [0.0, 0.0, -d*v_m*a**2, d*v_m]
            ]
        )

        # Compute source term
        S = np.asarray(
            [
                -d*v_m/(b*r)*diff_br,
                d*v_t**2/r*math.sind(phi) - 2*tau_w/b*math.cosd(alpha),
                -d*v_t*v_m/r*math.sind(phi)-2*tau_w/b*math.sind(alpha),
                2*(tau_w*v + q_w)/(b*dedp_d)
            ]
        ) 

        dy = scipy.linalg.solve(M, S)

        out = {**velocity_triangle,
               **static_properties,
               **utils.add_string_to_keys(stagnation_properties, "0")}

        return dy, out
    
    solution = scipy.integrate.solve_ivp(
        lambda t,y: odefun(t,y)[0],
        [0, m_out],
        [v_m_in, v_t_in, d_in, p_in],
        method = "RK45",
        rtol = 1e-6,
        atol = 1e-6,
    )

    dy, exit_plane = odefun(solution.t[-1], solution.y[:, -1])
    results = {
        "inlet_plane": input,
        "exit_plane" : exit_plane}

    return results, {}

def r_fun(r_in, phi, m):

    "Calculate the radius from the meridonial coordinate"

    return r_in + math.sind(phi)*m

def b_fun(b_in, div, m):

    "Calculate the channel width from the meridonial coordinate"

    return b_in + 2*math.tand(div)*m

def evaluate_volute(independents, input, boundary_conditions, geometry, fluid, model_options, reference_values):

    """
    Evaluate the volute. This include the scroll and concial diffuser.

    Assumptions:
    - Single velocity component at volute exit
    - Incompressible flow in diffuser cone

    """

    # Load reference_values
    v_max = reference_values["v_max"]
    s_range = reference_values["s_range"]
    s_min = reference_values["s_min"]

    # Load independent variables
    v_out = independents["v_out_volute"]*v_max
    s_out = independents["s_out_volute"]*s_range + s_min

    # Load input variables
    h0_in = input["h0"]
    s_in = input["s"]

    # Load diffuser geometry
    A_out = geometry["area_out"]
    A_scroll = geometry["area_scroll"]

    # Load boundary conditions
    mass_flow_rate = boundary_conditions["mass_flow_rate"]

    # Load model options
    loss_model = model_options["loss_model"]["volute"]

    # Define velocity triangle
    velocity_triangle_out = {"v" : v_out}

    # Calculate thermophysical properties
    h0_out = h0_in
    h_out = h0_out - 0.5*v_out**2
    static_properties_out = fluid.get_props(cp.HmassSmass_INPUTS, h_out, s_out)
    stagnation_properties_out = fluid.get_props(cp.HmassSmass_INPUTS, h0_out, s_out)
    isentropic_stagnation_properties_out = fluid.get_props(cp.PSmass_INPUTS, stagnation_properties_out["p"], s_in)
    isentropic_static_properties_out = fluid.get_props(cp.PSmass_INPUTS, static_properties_out["p"], s_in)
    isentropic_properties_out = {**isentropic_static_properties_out,**utils.add_string_to_keys(isentropic_stagnation_properties_out, "0")}

    # Calculate mass flow rate and mach 
    m_out = static_properties_out["d"]*v_out*A_out
    Ma_out = v_out/static_properties_out["a"]

    # Evaluate scroll velocity (Assume incompressible flow in diffuser cone)
    v_scroll = v_out*A_out/A_scroll

    # Collect exit plane results
    exit_plane = {**velocity_triangle_out,
                   **static_properties_out,
                   **utils.add_string_to_keys(stagnation_properties_out, "0"),
                   "Ma" : Ma_out,
                   }

    # Evaluate losses 
    loss_input = {
        "inlet_plane" : input,
        "exit_plane" : exit_plane,
        "geometry" : geometry,
        "isentropic" : isentropic_properties_out,
        "factors" : model_options["factors"],
        "boundary_conditions" : boundary_conditions, 
    }
    loss_dict = lm.evaluate_loss_model(loss_model, loss_input, "volute")
    exit_plane = {**exit_plane, **loss_dict}

    # Evaluate residuals
    resiudals = {"mass_flow_out" : (m_out-mass_flow_rate)/mass_flow_rate,
                 "losses" : loss_dict["loss_error"],
    }

    # Store diffuser results
    results = {"inlet_plane" : input,
               "exit_plane": exit_plane}
    
    return results, resiudals
    

def compute_overall_performance(results, fluid, geometry):
    """
    Calculate the overall performance metrics of the centrifugal compressor.

    This function extracts necessary values from the `results` dictionary, performs calculations to determine
    overall performance metrics, and returns these in a dictionary.

    Parameters
    ----------
    results : dict
        A dictionary containing all necessary information, such as geometry and flow characteristics.

    Returns
    -------
    dict
        An dictionary containing the calculated performance metrics.
    """

    # Load boundary conditions
    p0_in = results["boundary_conditions"]["p0_in"]
    T0_in = results["boundary_conditions"]["T0_in"]
    omega = results["boundary_conditions"]["omega"]
    mass_flow_rate = results["boundary_conditions"]["mass_flow_rate"]

    # Get results for first and final plane
    first_plane = results["planes"].loc[0].to_dict()
    final_plane = results["planes"].loc[len(results["planes"])-1].to_dict()

    # Get impeller results
    impeller = results["impeller"]

    # Calculate enthalpy for isentropic compression
    stagnation_properties = fluid.get_props(cp.PSmass_INPUTS, final_plane["p0"], first_plane["s"])
    h0_out_is = stagnation_properties["h"]
    static_properties = fluid.get_props(cp.PSmass_INPUTS, final_plane["p"], first_plane["s"])
    h_out_is = static_properties["h"]
 
    # Calculate overall performance characteristics
    PR_tt = final_plane["p0"]/p0_in
    PR_ts = final_plane["p"]/p0_in
    power = mass_flow_rate*(final_plane["h0"]-first_plane["h0"])
    torque = mass_flow_rate*(abs(geometry["impeller"]["radius_out"]*impeller["exit_plane"]["v_t"]) - abs(geometry["impeller"]["radius_mean_in"]*impeller["inlet_plane"]["v_t"]))
    efficiency_ts = mass_flow_rate*(h_out_is - first_plane["h0"])/(omega*torque)*100
    efficiency_tt = mass_flow_rate*(h0_out_is - first_plane["h0"])/(omega*torque)*100

    # Calculate nondimensional parameters
    gamma = first_plane["gamma"]
    R = first_plane["cp"] - first_plane["cv"]
    m_non_dim = mass_flow_rate*np.sqrt(T0_in)/p0_in*np.sqrt(R/gamma)/(2*geometry["impeller"]["radius_out"])**2
    N_non_dim = omega*60/2/np.pi/np.sqrt(T0_in)/np.sqrt(gamma*R)*2*geometry["impeller"]["radius_out"]

    # Store all variables in dictionary
    overall = {
        "PR_tt": PR_tt,
        "PR_ts": PR_ts,
        "efficiency_tt": efficiency_tt,
        "efficiency_ts": efficiency_ts,
        "power": power,
        "torque": torque,
        "m_non_dim" : m_non_dim,
        "N_non_dim" : N_non_dim,
    }

    return overall





import numpy as np
from .. import math

def compute_losses(input_parameters, component):

    # Map loss model functions
    component_function = {
        "impeller" : compute_impeller_losses,
        "vaneless_diffuser" : compute_vaneless_diffuser_losses,
        "vaned_diffuser" : compute_vaned_diffuser_losses,
        "volute" : compute_volute_losses,
    }   

    loss_dict = component_function[component](input_parameters)

    return loss_dict

def compute_vaned_diffuser_losses(input_parameters):
    
    """
    Conrad (See Meroni paper)
    """

    # Load parameters
    r_in = input_parameters["geometry"]["radius_in"]
    r_out = input_parameters["geometry"]["radius_out"]
    theta_in = input_parameters["geometry"]["leading_edge_angle"]
    theta_out = input_parameters["geometry"]["trailing_edge_angle"]
    b_out = input_parameters["geometry"]["width_out"]
    b_throat = input_parameters["geometry"]["width_throat"]
    opening = input_parameters["geometry"]["opening"]
    pitch = input_parameters["geometry"]["pitch"]
    v_in = input_parameters["inlet_plane"]["v"]
    v_out = input_parameters["exit_plane"]["v"]
    v_m_out = input_parameters["exit_plane"]["v_m"]
    alpha_out = input_parameters["exit_plane"]["alpha"]
    Cf = input_parameters["factors"]["skin_friction"]

    # Skin friction
    L_b = (r_out-r_in)/math.cosd((theta_in+theta_out)/2)
    D_h = opening*b_throat/(opening+b_throat) + pitch*b_out/(pitch+b_out) 
    Y_sf = 2*Cf*v_m_out**2*L_b/D_h    

    # Incidence
    alpha_des = theta_out
    Y_inc = 0.6*math.sind(abs(alpha_out-alpha_des))*0.5*v_out**2

    # Total losses
    scale = 0.5*v_in**2
    Y_inc = Y_inc/scale
    Y_sf = Y_sf/scale

    # Store losses in dict
    loss_dict = {"incidence" : Y_inc,
                 "skin_friction" : Y_sf,
                 "loss_total" : Y_inc + Y_sf
                 }

    return loss_dict

def compute_volute_losses(input_parameters):

    """
    Roberto's volute losses
    """
    
    # Load parameters
    alpha_in = input_parameters["inlet_plane"]["alpha"]
    A_in = input_parameters["geometry"]["area_in"]
    A_scroll = input_parameters["geometry"]["area_scroll"]
    A_out = input_parameters["geometry"]["area_out"]

    # Calculate dissipation of radial kinetic energy
    zeta_rad = math.cosd(alpha_in)**2

    # Calculate dissipation of tangential kinetic energy due to expansion loss in scroll
    zeta_scroll = math.sind(alpha_in)**2*(1-A_in*math.sind(alpha_in)/A_scroll)**2

    # Calculate dissipation of tangential kinetic energy due to expansion loss in cone
    zeta_cone = math.cosd(alpha_in)**2*(A_in/A_scroll)**2*(1-A_scroll/A_out)**2

    # Store losses in dict
    loss_dict = {"loss_radial" : zeta_rad,
                 "loss_scroll" : zeta_scroll,
                 "loss_cone" : zeta_cone,
                 "loss_total" : zeta_rad + zeta_scroll + zeta_cone}

    return loss_dict


def compute_vaneless_diffuser_losses(input_parameters):

    """
    Roberto's vanless diffuser losses
    """
    
    # Load parameters
    Cf = input_parameters["factors"]["skin_friction"]
    r_out = input_parameters["geometry"]["radius_out"]
    r_in = input_parameters["geometry"]["radius_in"]
    b_in = input_parameters["geometry"]["width_in"]

    # Store losses in dict
    loss_dict = {
        "loss_total" : 2*Cf*(r_out - r_in)/b_in
    }

    return loss_dict

def compute_impeller_losses(input_parameters):

    """
    Oh. Impeller losses
    """
    
    # Load inlet plane
    v_in = input_parameters["inlet_plane"]["v"]
    v_m_in = input_parameters["inlet_plane"]["v_m"]
    w_in = input_parameters["inlet_plane"]["w"]
    w_m_in = input_parameters["inlet_plane"]["w_m"]
    w_t_in = input_parameters["inlet_plane"]["w_t"]
    beta_in = input_parameters["inlet_plane"]["beta"]
    d_in = input_parameters["inlet_plane"]["d"]
    h0_in = input_parameters["inlet_plane"]["h0"]
    a_in = input_parameters["inlet_plane"]["a"]
    mu_in = input_parameters["inlet_plane"]["mu"]

    # Load exit plane
    v_out = input_parameters["exit_plane"]["v"]
    v_t_out = input_parameters["exit_plane"]["v_t"]
    alpha_out = input_parameters["exit_plane"]["alpha"]
    beta_out = input_parameters["exit_plane"]["beta"]
    w_out = input_parameters["exit_plane"]["w"]
    u_out = input_parameters["exit_plane"]["blade_speed"]
    d_out = input_parameters["exit_plane"]["d"]
    h0_out = input_parameters["exit_plane"]["h0"]
    a_out = input_parameters["exit_plane"]["a"]

    # Load geometry
    t_cl = input_parameters["geometry"]["tip_clearance"]
    b_out = input_parameters["geometry"]["width_out"]
    L_ax = input_parameters["geometry"]["impeller_length"]
    z = input_parameters["geometry"]["number_of_blades"]
    r_in_hub = input_parameters["geometry"]["radius_hub_in"]
    r_in_tip = input_parameters["geometry"]["radius_tip_in"]
    r_out = input_parameters["geometry"]["radius_out"]
    A_in = input_parameters["geometry"]["area_in"]
    A_throat = input_parameters["geometry"]["area_throat"]

    # Load model factors
    # Cf = input_parameters["factors"]["skin_friction"] 
    wake_width = input_parameters["factors"]["wake_width"]

    # Incidence loss
    f_inc = 0.6
    Y_inc = f_inc*w_t_in**2/2 

    # Blade loading loss
    Df = 1 - w_out/w_in + 0.75*(h0_out - h0_in)*w_out/((z/np.pi*(1-r_in_tip/r_out)+2*r_in_tip/r_out)*w_in*u_out**2)
    Y_bld = 0.05*Df**2*u_out**2 

    # Skin friction loss
    L_b = np.pi/8*(r_out*2-(r_in_tip + r_in_hub)-b_out+2*L_ax)*(2/(math.cosd(beta_in)+math.cosd(beta_out)))
    D_h = 2*r_out*math.cosd(beta_out)/(z/np.pi + 2*r_out*math.cosd(beta_out)/b_out) + 0.5*(r_in_tip+r_in_hub)/r_out*math.cosd(beta_in)/(z/np.pi + (r_in_tip + r_in_hub)/(r_in_tip-r_in_hub)*math.cosd(beta_in))
    w_avg = (v_in + v_out + 3*w_in + 3*w_out)/8
    Re = u_out*D_h*d_in/mu_in
    Cf = 0.0412*Re**(-0.1925)
    Y_sf = 2*Cf*L_b/D_h*w_avg**2

    # Clearance loss
    Y_cl = 0.6*t_cl/b_out*v_t_out*(4*np.pi/(b_out*z)*(r_in_tip**2 - r_in_hub**2)/((r_out - r_in_tip)*(1+d_out/d_in))*v_t_out*v_m_in)**(0.5)

    # Mixing loss
    width_diffuser = b_out # Assume no change in width from impeller exit to diffuser inlet
    b_star = width_diffuser/b_out
    Y_mix = 1/(1+math.tand(alpha_out)**2)*((1-wake_width-b_star)/(1-wake_width))**2*v_out**2/2

    # Shock loss models
    w_th = w_m_in*A_in/A_throat
    Y_shock = 0.56*max(0,((w_th/a_in)**2-1)**3)

    # Total losses
    scale = 0.5*v_in**2
    Y_inc = Y_inc/scale
    Y_bld = Y_bld/scale
    Y_sf = Y_sf/scale
    Y_cl = Y_cl/scale
    Y_mix = Y_mix/scale
    Y_tot = Y_inc + Y_bld + Y_sf + Y_cl + Y_mix + Y_shock

    # Scale loss coefficent
    loss_dict = {"incidence" : Y_inc,
                 "blade_loading" : Y_bld,
                 "skin_friction" : Y_sf,
                 "tip_clearance" : Y_cl,
                 "wake_mixing" : Y_mix,
                 "shock" : Y_shock,
                 "loss_total" : Y_tot}
    
    return loss_dict

def compute_parasitic_losses(input_parameters):

    """
    Oh's parasitic losses
    """

    # Load boundary conditions 
    mass_flow_rate = input_parameters["boundary_conditions"]["mass_flow_rate"]

    # Load inlet plane
    v_t_in = input_parameters["inlet_plane"]["v_t"]
    w_in = input_parameters["inlet_plane"]["w"]
    d_in = input_parameters["inlet_plane"]["d"]
    h0_in = input_parameters["inlet_plane"]["h0"]

    # Load exit plane
    v_t_out = input_parameters["exit_plane"]["v_t"]
    alpha_out = input_parameters["exit_plane"]["alpha"]
    w_out = input_parameters["exit_plane"]["w"]
    u_out = input_parameters["exit_plane"]["blade_speed"]
    d_out = input_parameters["exit_plane"]["d"]
    h0_out = input_parameters["exit_plane"]["h0"]
    mu_out = input_parameters["exit_plane"]["mu"]

    # Load geometry
    t_cl = input_parameters["geometry"]["tip_clearance"]
    b_out = input_parameters["geometry"]["width_out"]
    b_in = input_parameters["geometry"]["width_in"]
    L_ax = input_parameters["geometry"]["impeller_length"]
    z = input_parameters["geometry"]["number_of_blades"]
    r_in_tip = input_parameters["geometry"]["radius_tip_in"]
    r_out = input_parameters["geometry"]["radius_out"]
    r_mean_in = input_parameters["geometry"]["radius_mean_in"]

    # Recirculation losses
    Df = 1 - w_out/w_in + 0.75*(h0_out - h0_in)*w_out/((z/np.pi*(1-r_in_tip/r_out)+2*r_in_tip/r_out)*w_in*u_out**2)
    Y_rc = 8e-5*np.sinh(3.5*(alpha_out*np.pi/180)**3)*Df**2*u_out**2

    # Disk friction losses
    Re = d_out*u_out*r_out/mu_out
    f_df = 2.67/Re**0.5*(Re<3e5) + 0.0622/Re**0.2*(Re>=3e5)
    d_avg = (d_in + d_out)/2
    Y_df = f_df*d_avg*r_out**2*u_out**3/(f*mass_flow_rate)

    # Leakage loss
    r_avg = (r_mean_in + r_out)/2
    b_avg = (b_in + b_out)/2 
    delta_p = mass_flow_rate*(r_out*v_t_out - r_mean_in*v_t_in)/(z*r_avg*b_avg*L_m)
    u_cl = 0.816*np.sqrt(2*delta_p/d_out)
    m_cl = d_out*z*t_cl*L_m*u_cl
    Y_lk = m_cl*u_cl*u_out/(2*mass_flow_rate)

    loss_dict = {"recirculation" : Y_rc,
                 "disk_friction" : Y_df,
                 "leakage" : Y_lk,
                 "loss_total" : Y_rc + Y_df + Y_lk,
                 }
    
    return loss_dict



import numpy as np
from .. import math
from .. import utilities as utils

def calculate_full_geometry(geometry):

    r"""
    Computes the complete geometry of a centrifugal compressor based on input geometric parameters.

    Parameters
    ----------
    geometry : dict
        A dictionary containing the input geometry parameters

    Returns
    -------
    dict
        A dictionary with both the original and newly computed geometry parameters.
    """

    # Create a copy of the input dictionary to avoid mutating the original
    for key in geometry.keys():
        if key == "impeller":
            imp_geometry = calculate_impeller_geoemtry(geometry["impeller"])
            geometry["impeller"] = imp_geometry
        elif key == "vaneless_diffuser":
            vld_geometry = calculate_vaneless_diffuser_geometry(geometry["vaneless_diffuser"])
            geometry["vaneless_diffuser"] = vld_geometry
        elif key == "vaned_diffuser":
            vd_geometry = calculate_vaned_diffuser_geometry(geometry["vaned_diffuser"])
            geometry["vaned_diffuser"] = vd_geometry
        elif key == "volute":
            vol_geometry = calculate_volute_geometry(geometry["volute"])
            geometry["volute"] = vol_geometry
         
    return geometry

def calculate_impeller_geoemtry(imp_geometry):

    # Extract initial radii and compute throat values
    radius_hub_in = imp_geometry["radius_hub_in"] # Hub radius at inlet
    radius_tip_in = imp_geometry["radius_tip_in"] # Tip radius at inlet
    radius_out = imp_geometry["radius_out"] # Radius of impeller exit
    width_out = imp_geometry["width_out"] # Blade with at impeller exit
    area_throat_ratio = imp_geometry["area_throat_ratio"]

    # Calculate geometry
    radius_mean_in = (radius_tip_in + radius_hub_in) / 2 # Mean radius at inlet
    width_in = radius_tip_in - radius_hub_in
    A_in = np.pi * (radius_tip_in**2 - radius_hub_in**2) # Area inlet
    A_out = 2*np.pi*radius_out*width_out # Area exit

    new_parameters = {"radius_mean_in" : radius_mean_in,
                      "width_in" : width_in,
                      "area_in" : A_in,
                      "area_out" : A_out,
                      "area_throat" : A_in*area_throat_ratio}

    return {**imp_geometry, **new_parameters} 

def calculate_vaneless_diffuser_geometry(vld_geometry):

    # Load geometry
    b_in = vld_geometry["width_in"] # Channel width in
    r_in = vld_geometry["radius_in"] # Radius in
    r_out = vld_geometry["radius_out"] # Radius out
    b_out = vld_geometry["width_out"] # Wall divergence semi-angle    
    # div = vld_geometry["wall_divergence"] # Wall divergence semi-angle


    # Calculate geometry
    # b_out = b_in + 2*math.tand(div)*(r_out - r_in) # Channel width out
    div = math.arctand((0.5*(b_in-b_out)/(r_out-r_in)))
    A_in = 2*np.pi*r_in*b_in # Area in
    A_out = 2*np.pi*r_out*b_out # Area out

    new_parameters = {
        "area_in" : A_in,
        "area_out" : A_out,
        # "width_out" : b_out,
        "wall_divergence" : div,
        }

    return {**vld_geometry, **new_parameters} 

def calculate_vaned_diffuser_geometry(vd_geometry):

    # Load geometry
    r_out = vd_geometry["radius_out"] # Radius out
    r_in = vd_geometry["radius_in"] # Radius in
    b_in = vd_geometry["width_in"] # Channel width in
    b_out = vd_geometry["width_out"] # Channel width out
    theta_in = vd_geometry["leading_edge_angle"] # Leading edge blade angle
    theta_out = vd_geometry["trailing_edge_angle"] # Trailing edge blade angle
    z = vd_geometry["number_of_vanes"] # Number of vanes
    throat_location_factor = vd_geometry["throat_location_factor"]
    area_throat_ratio = vd_geometry["area_throat_ratio"]

    # Calculate geometry
    A_in = 2*np.pi*r_in*b_in # Area in
    A_out = 2*np.pi*r_out*b_out # Area out
    theta_avg = (theta_in + theta_out)/2 # Mean blade angle
    camber_angle = abs(theta_in - theta_out) # Camber angle
    solidity = z*(r_out -r_in)/(2*np.pi*r_in*math.cosd(theta_avg)) # Solidity 
    loc_camber_max = (2-abs(theta_in - theta_avg)/camber_angle)/3 # Location of max camber
    pitch = 2*np.pi*r_out/z
    b_throat = b_in*(1-throat_location_factor) + b_out*throat_location_factor

    new_parameters = {"area_out" : A_out,
                      "area_in" : A_in,
                      "theta_mean" : theta_avg,
                      "camber_angle" : camber_angle,
                      "solidity" : solidity,
                      "loc_camber_max" : loc_camber_max,
                      "pitch" : pitch,
                      "width_throat" : b_throat,
                      "area_throat" : A_in*area_throat_ratio,
                      }

    return {**vd_geometry, **new_parameters} 

def calculate_volute_geometry(vol_geometry):

    # Load geometry
    r_out = vol_geometry["radius_out"] # Radius out
    r_in = vol_geometry["radius_in"] # Radius in
    r_scroll = vol_geometry["radius_scroll"] # Radius at scroll exit
    width_in = vol_geometry["width_in"] # Width at scroll inlet

    # Calculate geometry
    A_in = 2*np.pi*r_in*width_in # Area at scroll inlet
    A_out = np.pi*r_out**2 # Area at cone diffuser exit
    A_scroll = np.pi*r_scroll**2 # Area at scroll exit

    new_parameters = {"area_out" : A_out,
                      "area_in" : A_in,
                      "area_scroll" : A_scroll}

    return {**vol_geometry, **new_parameters} 


def validate_compressor_geometry(geom, display=False):
    """
    Performs validation of a centrifugal compressor's geometry configuration.

    - Ensures all required geometry parameters are specified in 'geom'.
    - Verifies that no parameters beyond the required set are included.
    - Asserts that all geometry parameter values are numeric (integers or floats).
    - Ensures non-angle parameters contain only non-negative values.
    - Angle parameters can be negative.

    This function is intended as a preliminary check before in-depth geometry analysis.

    Parameters
    ----------
    geom : dict
        The geometry configuration parameters for the turbine as a dictionary.

    Returns
    -------
    bool
        True if all validations pass, indicating a correctly structured geometry configuration.
    """

    # Define the list of keys that must be defined in the configuration file
    required_keys = {
        "radius_hub_in",
        "radius_tip_out",
        "radius_out",
        "impeller_width_out",
    }

    # Angular variables need not be non-negative
    angle_keys = []

    # Check that there are no missing or extra variables
    utils.validate_keys(geom, required_keys, required_keys)

    # Check array size and values
    for key, value in geom.items():

        # Check that all variables are numpy arrays
        if not isinstance(value, np.ndarray):
            raise TypeError(f"Parameter '{key}' must be a NumPy array.")

        # Check that values of parameters are numeric
        if not math.all_numeric(value):
            raise ValueError(
                f"Parameter '{key}' must be an array of numeric types. Current value: {value}"
            )

        # Check that non-angle variables are not negative
        if key not in angle_keys and not math.all_non_negative(value):
            raise ValueError(
                f"Parameter '{key}' must be an array of non-negative numbers. Current value: {value}"
            )

    msg = "The structure of the turbine geometry parameters is valid"
    if display:
        print(msg)

    return msg






import numpy as np

# List of valid options
SLIP_MODELS = [
    "no-slip",
    "wiesner",
]

def evaluate_slip(slip_model, input, geometry):

    # Function mappings for each slip model
    slip_model_functions = {
        SLIP_MODELS[0] : lambda x, y: 0,
        SLIP_MODELS[1] : get_slip_weisner,
    }

    # Evaluate slip model
    if slip_model in SLIP_MODELS:
        slip_velocity = slip_model_functions[slip_model](input, geometry)
        return slip_velocity
    else:
        options = ", ".join(f"'{k}'" for k in SLIP_MODELS)
        raise ValueError(
            f"Invalid slip model: '{slip_model}'. Available options: {options}"
        )

def get_slip_weisner(input, geometry):

    # Load variables
    u_out = input["u_out"]
    theta_out = geometry["trailing_edge_angle"]
    z = geometry["number_of_blades"]

    return u_out*np.sqrt(np.cos(theta_out))/(z**0.7)



from pydantic import (
    BaseModel,
    Field,
    model_validator,
    field_validator,
    ConfigDict,
    root_validator,
)
from typing import Optional, List, Union, Literal, Tuple, Dict
from typing_extensions import Annotated
from enum import Enum

from ..centrifugal_compressor import (
    LOSS_MODELS,
    LOSS_COEFFICIENTS,
    SLIP_MODELS,
    CHOKING_CRITERIONS,
)
from ..pysolver_view import (
    SOLVER_OPTIONS,
    VALID_LIBRARIES_AND_METHODS,
    DERIVATIVE_METHODS,
)


class OperationPoints(BaseModel):
    """
    Model describing the operational points of an axial turbine.

    Attributes
    ----------
    fluid_name : str
        Name of the fluid.
    T0_in : float
        Inlet total temperature (must be greater than 0).
    p0_in : float
        Inlet total pressure (must be greater than 0).
    mass_flow_rate : float
        Mass flow rate (must be greater than 0).
    omega : float
        Rotational speed (must be greater than 0).
    alpha_in : float
        Inlet angle.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid")
    fluid_name: str
    T0_in: Annotated[float, Field(gt=0)]
    p0_in: Annotated[float, Field(gt=0)]
    mass_flow_rate: Annotated[float, Field(gt=0)]
    omega: Annotated[float, Field(gt=0)]
    alpha_in: float

LossModelEnum = Enum('LossModels', dict(zip([model.upper() for model in LOSS_MODELS], LOSS_MODELS)))
LossCoefficientEnum = Enum('LossCoefficients', dict(zip([model.upper() for model in LOSS_COEFFICIENTS], LOSS_COEFFICIENTS)))

# Define the basic loss configuration for each component
class LossConfig1(BaseModel):
    model_config = ConfigDict(extra="forbid", use_enum_values=True)
    model: LossModelEnum = "isentropic"
    loss_coefficient: LossCoefficientEnum = "static_enthalpy_loss"

class LossConfig2(BaseModel):
    model_config = ConfigDict(extra="forbid", use_enum_values=True)
    impeller: LossConfig1 = LossConfig1()
    vaneless_diffuser: LossConfig1 = LossConfig1()
    vaned_diffuser: LossConfig1 = LossConfig1()
    volute: LossConfig1 = LossConfig1()

class Factors(BaseModel):

    model_config = ConfigDict(extra="forbid", use_enum_values=True)
    skin_friction : Annotated[float, Field(ge=0)] = 0.00
    wall_heat_flux : Annotated[float, Field(ge=0)] = 0.00
    wake_width : Annotated[float, Field(gt=0)] = 0.366

SlipModelEnum = Enum('SlipModels', dict(zip([model.upper() for model in SLIP_MODELS], SLIP_MODELS)))
ChokingCriterionEnum = Enum('ChokingCriterionModels', dict(zip([model.upper() for model in CHOKING_CRITERIONS], CHOKING_CRITERIONS)))


class SimulationOptions(BaseModel):
    """
    Model describing simulation options.

    Attributes
    ----------
    slip_model : str
        The type of slip model. Should be one of the predefined slip models.
    loss_model : LossModel
        Object describing loss model setup and related parameters.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid", use_enum_values=True)
    slip_model : SlipModelEnum
    loss_model: Union[LossConfig1, LossConfig2]
    vaneless_diffuser_model : Literal["algebraic", "1D"]
    factors : Factors
    rel_step_fd : Annotated[float, Field(gt=0)] = 1e-4
    choking_criterion : ChokingCriterionEnum 

class Impeller(BaseModel):

    model_config = ConfigDict(extra="forbid", use_enum_values=True)
    radius_hub_in: Annotated[float, Field(gt=0)] 
    radius_tip_in: Annotated[float, Field(gt=0)]
    radius_out : Annotated[float, Field(gt=0)]
    width_out : Annotated[float, Field(gt=0)]
    leading_edge_angle : Annotated[float, Field(le=90), Field(gt=-90)]
    trailing_edge_angle : Annotated[float, Field(le=90), Field(gt=-90)]
    number_of_blades : Annotated[float, Field(gt=0)]
    length_axial : Annotated[float, Field(gt=0)]
    length_meridional : Annotated[float, Field(gt=0)]
    tip_clearance : Annotated[float, Field(gt=0)]
    area_throat_ratio : Annotated[float, Field(gt=0)]

class VanedDiffuser(BaseModel):

    model_config = ConfigDict(extra="forbid", use_enum_values=True)
    radius_out : Annotated[float, Field(gt=0)]
    radius_in : Annotated[float, Field(gt=0)]
    width_in : Annotated[float, Field(gt=0)]
    width_out : Annotated[float, Field(gt=0)]
    leading_edge_angle : Annotated[float, Field(le=90), Field(gt=-90)]
    trailing_edge_angle : Annotated[float, Field(le=90), Field(gt=-90)]
    number_of_vanes : Annotated[float, Field(gt=0)]
    opening : Annotated[float, Field(gt=0)]
    throat_location_factor : Annotated[float, Field(le=1), Field(gt=0)]
    area_throat_ratio : Annotated[float, Field(gt=0)]

class VanelessDiffuser(BaseModel):

    model_config = ConfigDict(extra="forbid", use_enum_values=True)
    # wall_divergence : Annotated[float, Field(le=90), Field(gt=-90)]
    radius_out : Annotated[float, Field(gt=0)]
    radius_in : Annotated[float, Field(gt=0)]
    width_in : Annotated[float, Field(gt=0)]
    width_out : Annotated[float, Field(gt=0)]

class Volute(BaseModel):

    model_config = ConfigDict(extra="forbid", use_enum_values=True)
    radius_out : Annotated[float, Field(gt=0)] 
    radius_in : Annotated[float, Field(gt=0)]
    radius_scroll : Annotated[float, Field(gt=0)]
    width_in : Annotated[float, Field(gt=0)]

class Geometry(BaseModel):

    model_config = ConfigDict(extra="forbid", use_enum_values=True)
    impeller : Impeller
    vaneless_diffuser : VanelessDiffuser = None
    vaned_diffuser : VanedDiffuser = None
    volute : Volute = None

    # Model validator to remove None fields after validation
    @model_validator(mode="after")
    def remove_none_fields(self):
        # Collect fields with None values to remove after iteration
        fields_to_remove = [field_name for field_name, value in self.__dict__.items() if value is None]
        
        # Delete attributes with None values
        for field_name in fields_to_remove:
            delattr(self, field_name)
            
        return self

class PerformanceMap(BaseModel):
    """
    Model describing a performance map for axial turbine analysis.

    Attributes
    ----------
    fluid_name : Union[str, List[str]]
        Name of the fluid or list of fluid names.
    T0_in : float, list
        Inlet total temperature or list of inlet total temperatures (each must be greater than 0).
    p0_in : float, list
        Inlet total pressure or list of inlet total pressures (each must be greater than 0).
    mass_flow_rate : float, list
        Mass flow rate or list of mass flow rates (each must be greater than 0).
    omega : float, list
        Rotational speed or list of rotational speeds (each must be greater than 0).
    alpha_in : float, list
        Inlet angle or list of inlet angles.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid")
    fluid_name: Union[str, List[str]]
    T0_in: Union[Annotated[float, Field(gt=0)], List[Annotated[float, Field(gt=0)]]]
    p0_in: Union[Annotated[float, Field(gt=0)], List[Annotated[float, Field(gt=0)]]]
    mass_flow_rate: Union[Annotated[float, Field(gt=0)], List[Annotated[float, Field(gt=0)]]]
    omega: Union[Annotated[float, Field(gt=0)], List[Annotated[float, Field(gt=0)]]]
    alpha_in: Union[float, List[float]]

SolverOptionEnum = Enum('SolverOptions', dict(zip([model.upper() for model in SOLVER_OPTIONS], SOLVER_OPTIONS)))
DerivativeMethodEnum = Enum('DerivativeMethods', dict(zip([model.upper() for model in DERIVATIVE_METHODS], DERIVATIVE_METHODS)))

class SolverOptions(BaseModel):

    """
    Model describing solver options for performance analysis.

    Attributes
    ----------
    method : str, optional
        The root solver method. Should be one of the predefined solver options. Default is 'hybr'.
    tolerance : float, optional
        Tolerance for convergence. Default is 1e-8.
    max_iterations : int, optional
        Maximum number of iterations. Default is 100.
    derivative_method : str, optional
        The method for finite difference approximation calculation. Should be one of the predefined derivative methods. Default is '2-point'.
    derivative_abs_step : float, optional
        Absolute step size for finite difference approximation calculation. Default is 1e-6.
    print_convergence : bool, optional
        Flag indicating whether to print convergence information. Default is True.
    plot_convergence : bool, optional
        Flag indicating whether to plot convergence. Default is False.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid", use_enum_values=True)
    method : SolverOptionEnum = SOLVER_OPTIONS[0]
    tolerance: float = 1e-8
    max_iterations: int = 100
    derivative_method : DerivativeMethodEnum = DERIVATIVE_METHODS[0]
    derivative_abs_step: float = 1e-6
    print_convergence: bool = True
    plot_convergence: bool = False

class PerformanceAnalysis(BaseModel):

    model_config = ConfigDict(extra="forbid")
    performance_map : PerformanceMap = None
    solver_options : SolverOptions = SolverOptions()
    initial_guess : Dict = {}

class CentrifugalCompressor(BaseModel):
    """
    Model describing an axial turbine.

    Attributes
    ----------
    turbomachinery : str
        The type of turbomachinery, which for this class is "centrifugal_compressor".
    operation_points : OperationPoints
        Object containing operation points data.
    simulation_options : SimulationOptions
        Object containing simulation options data.
    geometry : GeometryPerformanceAnalysis, optional
        Object containing geometry performance analysis data. Default is None.
    performance_analysis : PerformanceAnalysis, optional
        Object containing performance analysis data. Default is a new instance of PerformanceAnalysis().
    design_optimization : DesignOptimization, optional
        Object containing design optimization data. Default is a new instance of DesignOptimization().

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid")
    turbomachinery: Literal["centrifugal_compressor"]
    operation_points: Union[List[OperationPoints], OperationPoints]
    simulation_options: SimulationOptions
    geometry: Geometry
    performance_analysis: PerformanceAnalysis = PerformanceAnalysis()
    # design_optimization: DesignOptimization = DesignOptimization()



from .performance_analysis import *
# from .design_optimization import *
from .geometry_model import *
from .flow_model import *
from .loss_model import *
from .slip_model import *
from .choking_criterion import *



import os
import yaml
import copy
import time
import datetime
import itertools
import numpy as np
import pandas as pd
import CoolProp as cp
import matplotlib.pyplot as plt

from .. import math
from .. import pysolver_view as psv
from .. import utilities as utils
from .. import properties as props
from . import geometry_model as geom
from . import flow_model as flow
from scipy.stats import qmc
from scipy import optimize


SOLVER_MAP = {"lm": "Lavenberg-Marquardt", "hybr": "Powell's hybrid"}
"""
Available solvers for performance analysis.
"""

def compute_performance(
    config,
    operation_points,
    stop_on_failure=False,
    export_results = False,
    out_dir = "output",
    out_filename = None,
):
    # Get list of operating points
    operation_points = generate_operation_points(operation_points)

    # Initialize initial guess
    initial_guess = config["performance_analysis"]["initial_guess"]
    
    # Initialize lists to hold each solution and solver 
    solution_data = []
    solver_container = []

    # Calculate performance at all operation points
    print_operation_points(operation_points)
    for i, operation_point in enumerate(operation_points):
        print(f"\n Computing operation point {i+1} of {len(operation_points)}")
        print_boundary_conditions(operation_point)

        # Update initial guess
        if i > 0:
            closest_x, closest_index = find_closest_operation_point(
                operation_point,
                operation_points[:i],  # Use up to the previous point
                solution_data[:i],  # Use solutions up to the previous point
                solver_container[:i],
            )
            print(f" Using solution from point {closest_index+1} as initial guess")
            initial_guess = closest_x
        # try:
        # Compute performance
        solver = compute_single_operation_point(
                            operation_point,
                            initial_guess,
                            config["geometry"],
                            config["simulation_options"],
                            config["performance_analysis"]["solver_options"],
                            )

        # Collect solution and solver
        solution_data.append(solver.problem.vars_real)
        solver_container.append(solver)

        # except Exception as e:
        #     handle_failure(i, e, stop_on_failure)
        #     solver_container.append(None)
        #     solution_data.append([])

    # Print final report
    print_simulation_summary(solver_container)

    # Export results
    if export_results:
        export_results_excel(solver_container, out_dir, out_filename)

    return solver_container

def handle_failure(point_index, error, stop_on_failure):
    print(f" Computation of point {point_index} failed")
    print(f" Error: {error}")
    if stop_on_failure:
        raise Exception(error)

def export_results_excel(solver_container, out_dir, out_filename):

    """
    Export results to excel file
    Each component has a distinct sheet in the excel file
    """

    RESULTS_KEYS = ["overall", "boundary_conditions", "impeller", "vaneless_diffuser", "vaned_diffuser", "volute"]

    # Initialize results structures
    results = {key : [] for key in solver_container[0].problem.results.keys() if key in RESULTS_KEYS}
    solver_data = []
    
    # Loop through each solver result in the container
    for i, solver in enumerate(solver_container):
        if solver is None:
            # If solver failed, skip or add a placeholder
            for key in results.keys():
                results[key].append({})
            solver_data.append({})
            continue
        
        else:
            # Performance data
            for key in results.keys():
                val = solver.problem.results[key]
                # For components (impeller etc.)
                if isinstance(val, dict) and set(val.keys()).issuperset({"inlet_plane", "exit_plane"}):
                    results[key].append({
                        **utils.add_string_to_keys(val["inlet_plane"], "_in"), 
                        **utils.add_string_to_keys(val["exit_plane"], "_out"),
                        **(utils.add_string_to_keys(val["throat_plane"], "_throat") if "throat_plane" in val else {})
                    })
                
                # For entries with 
                elif isinstance(val, dict):
                    results[key].append(val)

            # Solver data
            solver_data.append({
                "completed": True,
                "success": solver.success,
                "message": solver.message,
                "grad_count": solver.convergence_history["grad_count"][-1],
                "func_count": solver.convergence_history["func_count"][-1],
                "func_count_total": solver.convergence_history["func_count_total"][-1],
                "norm_residual": solver.convergence_history["norm_residual"][-1],
                "norm_step": solver.convergence_history["norm_step"][-1],
            })



    # Convert lists to a DataFrame
    for key, val in results.items():
        results[key] = pd.DataFrame(results[key])
    results["solver"] = pd.DataFrame(solver_data)

    # Create a directory to save simulation results
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    # Define filename with unique date-time identifier
    if out_filename == None:
        current_time = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        out_filename = f"performance_analysis_{current_time}"

    filepath = os.path.join(out_dir, f"{out_filename}.xlsx")
    with pd.ExcelWriter(filepath, engine="openpyxl") as writer:
        for sheet_name, df in results.items():
            # Write each DataFrame to a specific sheet
            df.to_excel(writer, sheet_name=sheet_name, index=False)

    print(f"Results successfully exported to {filepath}")

    return

def compute_single_operation_point(
    operating_point,
    initial_guess,
    geometry,
    simulation_options,
    solver_options,
):

    # Initialize problem object
    problem = CentrifugalCompressorProblem(geometry, simulation_options)
    problem.update_boundary_conditions(operating_point)
    solver_options = copy.deepcopy(solver_options)

    # Get solver method array
    solver_methods = [solver_options["method"]] + [method for method in SOLVER_MAP.keys() if method != solver_options["method"]]

    # Solve operation point
    initial_guesses = problem.get_initial_guess(initial_guess)
    for initial_guess in initial_guesses:
        initial_guess_scaled = problem.scale_values(initial_guess)
        x0 = np.array(list(initial_guess_scaled.values()))
        problem.keys = list(initial_guess_scaled.keys())
        for method in solver_methods:
            solver_options["method"] = method
            solver = psv.NonlinearSystemSolver(problem, **solver_options)
            try:
                solver.solve(x0)
            except Exception as e:
                if solver.func_count == 0:
                    raise e 
                print(f" Error during solving: {e}")
                solver.success = False
            if solver.success:
                    break
            
        if solver.success:
                    break
    
    if not solver.success:
        print("WARNING: All attempts failed to converge")
        # TODO: Add messages to Log file

    return solver


def find_closest_operation_point(current_op_point, operation_points, solution_data, solver_container):
    """
    Find the solution vector and index of the closest operation point in the historical data.

    Parameters
    ----------
    current_op_point : dict
        The current operation point we want to compare.
    operation_points : list of dict
        A list of historical operation points to search through.
    solution_data : list
        A list of solution vectors corresponding to each operation point.
    solver_container : list
        A list of solver objects corresponding to each operation point.

    Returns
    -------
    tuple
        A tuple containing the closest solution vector and the one-based index of the closest operation point.

    """
    min_distance = float("inf")
    closest_point_x = None
    closest_index = None

    for i, op_point in enumerate(operation_points):
        distance = get_operation_point_distance(current_op_point, op_point)
        if distance < min_distance and solver_container[i].success:
            min_distance = distance
            closest_point_x = solution_data[i]
            closest_index = i

    return closest_point_x, closest_index


def get_operation_point_distance(point_1, point_2, delta=1e-8):
    """
    Calculate the normalized distance between two operation points, with special consideration
    for angle measurements and prevention of division by zero for very small values.

    Parameters
    ----------
    point_1 : dict
        First operation point with numeric values.
    point_2 : dict
        Second operation point with numeric values.
    delta : float, optional
        A small constant to prevent division by zero. Default is 1e-8.

    Returns
    -------
    float
        The calculated normalized distance.
    """
    deviation_array = []
    for key in point_1:
        if isinstance(point_1[key], (int, float)) and key in point_2:
            value_1 = point_1[key]
            value_2 = point_2[key]

            if key == "alpha_in":
                # Handle angle measurements with absolute scale normalization
                deviation = np.abs(value_1 - value_2) / 90
            else:
                # Compute the relative difference with protection against division by zero
                max_val = max(abs(value_1), abs(value_2), delta)
                deviation = abs(value_1 - value_2) / max_val

            deviation_array.append(deviation)

    # Calculate the two-norm of the deviations
    return np.linalg.norm(deviation_array)


def generate_operation_points(operation_points):
    """
    Generates a list of dictionaries representing all possible combinations of
    operation points from a given input. If the input is a dictionary with ranges,
    it generates all combinations; if it's a list or array, it directly returns the
    operation points.

    Parameters
    ----------
    - operation_points (dict or list/np.ndarray): A dictionary with parameter names as
      keys and lists of parameter values as values, or a precomputed list of
      operation points.

    Returns
    -------
    - operation_points (list of dict): A list of dictionaries, each representing
      a unique combination of parameters from the performance map.
    """
    # If input is already a list or array, return it as operation points
    if isinstance(operation_points, (list, np.ndarray)):
        return operation_points

    # If input is a dictionary, generate operation points by creating all combinations
    if isinstance(operation_points, dict):
        # Ensure all values in the performance_map are iterables
        performance_map = {k: utils.ensure_iterable(v) for k, v in operation_points.items()}

        # Reorder performance map keys so 'p0_in' and 'mass_flow_rate' are prioritized last
        priority_keys = ["p0_in", "mass_flow_rate"]
        other_keys = [k for k in performance_map.keys() if k not in priority_keys]
        keys_order = other_keys + priority_keys
        performance_map = {k: performance_map[k] for k in keys_order if k in performance_map}

        # Generate all combinations of operation points
        keys, values = zip(*performance_map.items())
        operation_points = [dict(zip(keys, combination)) for combination in itertools.product(*values)]
        
        return operation_points

    # Raise an error if input is neither a dict nor a list/array
    msg = "operation_points must be either a list of dicts or a dict with ranges."
    raise TypeError(msg)



def validate_operation_point(op_point):
    """
    Validates that an operation point has exactly the required fields:
    'fluid_name', 'p0_in', 'T0_in', 'mass_flow_rate', 'alpha_in', 'omega'.

    Parameters
    ----------
    op_point: dict
        A dictionary representing an operation point.

    Returns
    -------
    ValueError: If the dictionary does not contain the required fields or contains extra fields.
    """
    REQUIRED_FIELDS = {"fluid_name", "p0_in", "T0_in", "mass_flow_rate", "alpha_in", "omega"}
    fields = set(op_point.keys())
    if fields != REQUIRED_FIELDS:
        missing = REQUIRED_FIELDS - fields
        extra = fields - REQUIRED_FIELDS
        raise ValueError(
            f"Operation point validation error: "
            f"Missing fields: {missing}, Extra fields: {extra}"
        )


# ------------------------------------------------------------------------------------------ #
# ------------------------------------------------------------------------------------------ #
# ------------------------------------------------------------------------------------------ #

class CentrifugalCompressorProblem(psv.NonlinearSystemProblem):

    def __init__(self, geometry, simulation_options):
        
        # Process turbine geometry
        self.geometry = geom.calculate_full_geometry(geometry)

        # Initialize simulation options
        # Specify loss model for each component if not given
        self.model_options = simulation_options
        if set(simulation_options["loss_model"].keys()) == {"model", "loss_coefficient"}:
            loss_model = simulation_options["loss_model"]
            self.model_options["loss_model"] = {key : loss_model for key in self.geometry.keys()}

        # Initialize list of keys for the independent variables
        self.keys = []

    def residual(self, x):
        
        # Create dictionary of scaled variables
        self.vars_scaled = dict(zip(self.keys, x))

        # Create dictionary of real variables
        self.vars_real = self.scale_values(self.vars_scaled, to_normalized=False)
        
        # Evaluate centrifugal compressor
        results = flow.evaluate_centrifugal_compressor(
            self.vars_scaled,
            self.boundary_conditions,
            self.geometry,
            self.fluid,
            self.model_options,
            self.reference_values,
        )

        self.results = results

        return np.array(list(results["residuals"].values()))

    def update_boundary_conditions(self, operation_point):
        """
        Update the boundary conditions of the problem with the provided operation point.

        This method updates the boundary conditions attributes used to evaluate the turbine performance.
        It also initializes a Fluid object using the 'fluid_name' specified in the operation point.
        The method computes additional properties and reference values like stagnation properties at
        the inlet, exit static properties, spouting velocity, and reference mass flow rate.
        These are stored in the object's internal state for further use in calculations.

        Parameters
        ----------
        operation_point : dict
            A dictionary containing the boundary conditions defining the operation point. 

        Returns
        -------
        None
            This method does not return a value but updates the internal state of the object.

        """

        # Define current operating point
        self.boundary_conditions = operation_point

        # Initialize fluid object
        self.fluid = props.Fluid(operation_point["fluid_name"], exceptions=True)

        # Rename variables
        p0_in = operation_point["p0_in"]
        T0_in = operation_point["T0_in"]
        omega = operation_point["omega"]

        # Compute stagnation properties at inlet
        state_in_stag = self.fluid.get_props(cp.PT_INPUTS, p0_in, T0_in)
        h0_in = state_in_stag["h"]
        s_in = state_in_stag["s"]

        # Store the inlet stagnation (h,s) for the first plane
        self.boundary_conditions["h0_in"] = h0_in
        self.boundary_conditions["s_in"] = s_in

        # Calculate reference velocity
        u_out = omega*self.geometry["impeller"]["radius_out"]
        h0_max = h0_in + 0.5*u_out**2
        ref_props_min = self.fluid.get_props(cp.PSmass_INPUTS, p0_in/2, s_in)
        h_ref_min = ref_props_min["h"]
        v0 = np.sqrt(2*(h0_max - h_ref_min))

        # Calculate max entropy
        s_max = s_in + u_out**2/T0_in

        # Define reference_values
        self.reference_values = {
            "s_range": s_max - s_in,
            "s_min": s_in,
            "v_max": v0,
            "angle_range": 180,
            "angle_min": -90,
        }

        return
      
    def scale_values(self, variables, to_normalized=True):
        """
        Convert values between normalized and real values.

        Parameters
        ----------
        variables: dict
            A dictionary containing values to be scaled.
        to_real: bool
            If True, scale to real values; if False, scale to normalized values.

        Returns
        -------
        An array of values converted between scales.
        """

        # Load parameters
        v0 = self.reference_values["v_max"]
        s_range = self.reference_values["s_range"]
        s_min = self.reference_values["s_min"]
        angle_range = self.reference_values["angle_range"]
        angle_min = self.reference_values["angle_min"]

        # Define dictionary of scaled values
        scaled_variables = {}

        for key, val in variables.items():
            if key.startswith("v") or key.startswith("w"):
                scaled_variables[key] = val / v0 if to_normalized else val * v0
            elif key.startswith("s"):
                scaled_variables[key] = (
                    (val - s_min) / s_range if to_normalized else val * s_range + s_min
                )
            elif key.startswith("b") or key.startswith("a"):
                scaled_variables[key] = (
                    (val - angle_min) / angle_range
                    if to_normalized
                    else val * angle_range + angle_min
                )

        return scaled_variables
    
    def __getstate__(self):

        """
        This function is called when dumping object using pickle.
        This function ensures that attribute types that are not supported by pickle is reset.
        Every action in this function should correspond to an action in __setstate__ 
        """

        # Create a copy of the object's state dictionary
        state = self.__dict__.copy()
        # Remove the unpickleable 'fluid' entry
        state['fluid'] = None
        return state

    def __setstate__(self, state):

        """
        This function is called when loading a pickle file.
        This function ensures that attribute types that are not supported by pickle are restored.
        Every action in this function should correspond to an action in __getstate__ 
        """

        # Restore the attributes
        self.__dict__.update(state)

        # Recreate the 'fluid' attribute
        self.fluid = props.Fluid(self.boundary_conditions["fluid_name"])

    def get_initial_guess(self, initial_guess):

        """
        3 options:
            1. Give guess of efficiency_tt and phi_out. Initial guess calculated heuristically
            2. Give bounds of efficiency_tt and phi_out, and a sample size. Initial guess calculated by the best 
            3. Give complete set (real values)
        """

        valid_keys_1 = ["efficiency_impeller", "phi_impeller", "Ma_vaned_diffuser"]
        valid_keys_2 = ["efficiency_impeller", "phi_impeller", "Ma_vaned_diffuser", "n_samples"]
        check = []
        check.append(set(valid_keys_1) == set(list(initial_guess.keys())))
        check.append(set(valid_keys_2) == set(list(initial_guess.keys())))

        if check[0]:
            # Initial guess determined from heurustic guess, with given parameters
            if isinstance(initial_guess["efficiency_impeller"], (list, np.ndarray)):
                # Several initial guesses
                initial_guesses = []
                for i in range(len(initial_guess["efficiency_impeller"])):
                    guess = {key : val[i] for key, val in initial_guess.items()}
                    heuristic_guess = self.get_heuristic_guess(guess)
                    initial_guesses.append(heuristic_guess)
            else:
                # Single initial guess
                heuristic_guess = self.get_heuristic_guess(initial_guess)
                initial_guesses = [heuristic_guess]

        elif check[1]:
            # Generate initial guess using latin hypercube sampling
            bounds = [initial_guess["efficiency_impeller"], initial_guess["phi_impeller"], initial_guess["Ma_vaned_diffuser"]]

            n_samples = initial_guess["n_samples"]
            heuristic_inputs = latin_hypercube_sampling(bounds, n_samples)
            norm_residuals = np.array([])
            failures = 0
            for heuristic_input in heuristic_inputs:
                try:
                    guess = dict(zip(valid_keys_1, heuristic_input))
                    heuristic_guess = self.get_heuristic_guess(guess)
                    x = self.scale_values(heuristic_guess) # TODO: Add scaling
                    self.keys = x.keys() 
                    x0 = np.array(list(x.values()))
                    residual = self.residual(x0)
                    norm_residuals = np.append(norm_residuals, np.linalg.norm(residual))
                except:
                    failures += 1 
                    norm_residuals = np.append(norm_residuals, np.nan)

            print(f"Generating heuristic inital guesses from latin hypercube sampling")
            print(f"Number of failures: {failures} out of {n_samples} samples")
            print(f"Least norm of residuals: {np.nanmin(norm_residuals)}")
            heuristic_input = heuristic_inputs[np.nanargmin(norm_residuals)]
            initial_guess = dict(zip(valid_keys_1, heuristic_input))
            initial_guess = self.get_heuristic_guess(initial_guess)
            initial_guesses = [initial_guess]
        elif isinstance(initial_guess, dict):
            # Simply return the initial guess given
            initial_guesses = [initial_guess]
        else:
            raise ValueError("Provided initial guess is not valid.")
    
        return initial_guesses
    
    def get_heuristic_guess(self, guess):

        # Check vaneless diffuser model
        vaneless_model = self.model_options["vaneless_diffuser_model"]

        # initialize initial guess dictionary
        heuristic_guess = {}
        input = {}

        # Get initial guess
        for key in self.geometry.keys():
            guess_component = {var : val for var, val in guess.items() if var.endswith(key)}

            if key == "impeller":
                heuristic_guess_component, exit_state = self.get_impeller_guess(guess_component, input)
            elif key == "vaneless_diffuser" and vaneless_model == "algebraic":
                heuristic_guess_component, exit_state = self.get_vaneless_diffuser_guess(guess_component, input)
            elif key == "vaned_diffuser":
                heuristic_guess_component, exit_state = self.get_vaned_diffuser_guess(guess_component, input)
            elif key == "volute":
                heuristic_guess_component, exit_state = self.get_volute_guess(guess_component, input)

            # Store guess
            heuristic_guess = {**heuristic_guess, **heuristic_guess_component}

            # Prepare calculation of next component
            input = exit_state
        
        # Define keys to be deleted based on choking criterion
        choking_criterion = self.model_options["choking_criterion"]
        keys_to_delete = {
        "no_throat": {"w_throat_vaned_diffuser", "w_throat_impeller"},  
        "critical_isentropic_throat": {},  
        }
        # Remove specified keys for the selected choking criterion
        if choking_criterion in keys_to_delete:
            for key in keys_to_delete[choking_criterion]:
                heuristic_guess.pop(key, None)  # Remove key if it exists

        return heuristic_guess
    
    def get_impeller_guess(self, guess, input):

        # Load boundary conditions
        p0_in = self.boundary_conditions["p0_in"]
        T0_in = self.boundary_conditions["T0_in"]
        omega = self.boundary_conditions["omega"]
        mass_flow_rate = self.boundary_conditions["mass_flow_rate"]
        alpha_in = self.boundary_conditions["alpha_in"]

        # Load guess
        eta_tt = guess["efficiency_impeller"]
        phi_out = guess["phi_impeller"]
        
        # Load geometry
        z = self.geometry["impeller"]["number_of_blades"]
        theta_out = self.geometry["impeller"]["trailing_edge_angle"]
        r_out = self.geometry["impeller"]["radius_out"]
        r_in = self.geometry["impeller"]["radius_mean_in"]
        A_out = self.geometry["impeller"]["area_out"]
        A_in = self.geometry["impeller"]["area_in"]

        # Load velocity scale
        v0 = self.reference_values["v_max"]

        # Evaluate inlet thermpdynamic state
        stagnation_props_in = self.fluid.get_props(cp.PT_INPUTS, p0_in, T0_in)
        gamma = stagnation_props_in["cp"]/stagnation_props_in["cv"]
        h0_in = stagnation_props_in["h"]
        a0_in = stagnation_props_in["a"]
        s_in = stagnation_props_in["s"]

        # Approximate impeller exit total pressure
        u_out = r_out*omega
        Ma_out = u_out/a0_in
        slip_factor = 1-np.sqrt(math.cosd(theta_out))/(z**(0.7)*(1+phi_out*math.tand(theta_out)))
        p0_out = (1+(gamma-1)*eta_tt*slip_factor*(1+phi_out*math.tand(theta_out))*Ma_out**2)**(gamma/(gamma-1))*p0_in

        # Get impeller exit total enthalpy and entropy
        isentropic_props_out = self.fluid.get_props(cp.PSmass_INPUTS, p0_out, s_in)
        h0_out_is = isentropic_props_out["h"]
        h0_out = (h0_out_is-h0_in)/eta_tt + h0_in
        stagnation_props_out = self.fluid.get_props(cp.HmassP_INPUTS, h0_out, p0_out)
        s_out = stagnation_props_out["s"]

        # Get inlet velocity
        def calculate_inlet_residual(v_scaled):

            # Calculate velocity triangle
            u_in = r_in*omega
            v_in = v_scaled*v0
            v_m_in = v_in*math.cosd(alpha_in)
            v_t_in = v_in*math.sind(alpha_in)
            w_m_in = v_m_in
            w_t_in = v_t_in - u_in
            w_in = np.sqrt(w_m_in**2 + w_t_in**2)   
            beta_in = math.arctand(w_t_in/w_m_in)

            velocity_triangle_in = {"v_t" : v_t_in,
                    "v_m" : v_m_in,
                    "v" : v_in,
                    "alpha" : alpha_in,
                    "w_t" : w_t_in,
                    "w_m" : w_m_in,
                    "w" : w_in,
                    "beta" : beta_in,
                    "blade_speed" : u_in,
                    }

            # Calculate thermodynamic state
            h_in = h0_in - 0.5*v_in**2
            h0_rel_in = h_in + 0.5*w_in**2
            static_props_in = self.fluid.get_props(cp.HmassSmass_INPUTS, h_in, s_in)
            relative_props_in = self.fluid.get_props(cp.HmassSmass_INPUTS, h0_rel_in, s_in)

            # Calculate mass flow rate nad rothalpy
            m_in = static_props_in["d"]*v_m_in*A_in
            rothalpy_in = h0_rel_in - 0.5*u_in**2

            # Evaluate mass flow rate residual
            res = m_in - mass_flow_rate 

            check_in = {"mass_flow_rate" : m_in,
                        "rothalpy" : rothalpy_in,
                        "s_in" : s_in}

            return res, velocity_triangle_in, check_in
        
        def calculate_inlet_velocity():

            sol = optimize.root_scalar(lambda x: calculate_inlet_residual(x)[0], method = "secant", x0 = 0.2)
            delta_h, velocity_triangle_in, check_in  = calculate_inlet_residual(sol.root)
            return velocity_triangle_in, check_in

        velocity_triangle_in, check_in = calculate_inlet_velocity()

        # Calculate exit state
        v_m_out = phi_out*u_out
        v_t_out = slip_factor*(1+phi_out*math.tand(theta_out))*u_out
        v_out = np.sqrt(v_t_out**2 + v_m_out**2)
        alpha_out = math.arctand(v_t_out/v_m_out)
        w_m_out = v_m_out
        w_t_out = v_t_out - u_out
        w_out = np.sqrt(w_t_out**2 + w_m_out**2)
        beta_out = math.arctand(w_t_out/w_m_out)

        velocity_triangle_out = {"v_t" : v_t_out,
                    "v_m" : v_m_out,
                    "v" : v_out,
                    "alpha" : alpha_out,
                    "w_t" : w_t_out,
                    "w_m" : w_m_out,
                    "w" : w_out,
                    "beta" : beta_out,
                    "blade_speed" : u_out,
                    }
        
        # Get thermophysical properties
        h_out = h0_out -0.5*v_out**2
        h0_rel_out = h_out + 0.5*w_out**2
        static_props_out = self.fluid.get_props(cp.HmassSmass_INPUTS, h_out, s_out)

        # Calculate mass flow rate and rothalpy
        m_out = static_props_out["d"]*v_m_out*A_out
        rothalpy_out = h0_rel_out - 0.5*u_out**2

        exit_plane = {
            **velocity_triangle_out,
            **static_props_out,
        }

        # Store initial guess
        initial_guess = {
                    "v_in_impeller" : velocity_triangle_in["v"],
                    "w_throat_impeller" : velocity_triangle_in["w"]*0.5, # factor to ensure correct solution of thorat velocity
                    "w_out_impeller" : w_out,
                    "beta_out_impeller" : beta_out,
                    "s_out_impeller" : s_out,
                    }
        
        return initial_guess, exit_plane

    def get_vaneless_diffuser_guess(self, guess, input):
        """
        Guess constant alpha
        Simple correlation for static enthalpy loss coefficient
        """

        # Load input
        alpha_in  = input["alpha"]
        v_t_in = input["v_t"]
        v_in = input["v"]
        h_in = input["h"]
        d_in = input["d"]
        s_in = input["s"]

        # Load boundary conditions
        mass_flow_rate = self.boundary_conditions["mass_flow_rate"]

        # Load geometry
        r_out = self.geometry["vaneless_diffuser"]["radius_out"]
        r_in = self.geometry["vaneless_diffuser"]["radius_in"]
        b_in = self.geometry["vaneless_diffuser"]["width_in"]
        A_out = self.geometry["vaneless_diffuser"]["area_out"]

        # Load model options
        Cf = self.model_options["factors"]["skin_friction"]

        # Calculate exit velocity
        alpha_out = alpha_in
        # delta_M = np.exp(-Cf*(r_out-r_in)/(b_in*math.cosd(alpha_in)))
        delta_M = 1
        v_t_out = v_t_in*r_in/r_out*delta_M
        v_out = v_t_out/math.sind(alpha_out)
        v_m_out = v_out*math.cosd(alpha_out)
        velocity_triangle_out = {
            "v" : v_out,
            "v_t" : v_t_out,
            "v_m" : v_m_out,
            "alpha" : alpha_out
        }
        
        # Calculate exit entropy
        d_out = mass_flow_rate/(v_out*math.cosd(alpha_out)*A_out)
        h0_out = h_in + 0.5*v_in**2
        h_out = h0_out - 0.5*v_out**2
        static_props_out = self.fluid.get_props(cp.DmassHmass_INPUTS, d_out, h_out)
        s_out = static_props_out["s"]

        initial_guess = {
                    "v_out_vaneless_diffuser" : v_out,
                    "s_out_vaneless_diffuser" : s_out,
                    "alpha_out_vaneless_diffuser" : alpha_out,
                    }
        
        exit_state = {
            **velocity_triangle_out,
            **static_props_out,
        }

        return initial_guess, exit_state

    def get_vaned_diffuser_guess(self, guess, input):

        """
        Similar as vaneless diffuser
        Assume zero deviation at exit 
        """

        # Load guess
        Ma_out = guess["Ma_vaned_diffuser"]
        
        # Load input
        alpha_in  = input["alpha"]
        v_t_in = input["v_t"]
        v_in = input["v"]
        h_in = input["h"]
        a_in = input["a"]

        # Load boundary conditions
        mass_flow_rate = self.boundary_conditions["mass_flow_rate"]

        # Load geometry
        r_out = self.geometry["vaned_diffuser"]["radius_out"]
        r_in = self.geometry["vaned_diffuser"]["radius_in"]
        b_in = self.geometry["vaned_diffuser"]["width_in"]
        A_out = self.geometry["vaned_diffuser"]["area_out"]
        theta_out = self.geometry["vaned_diffuser"]["trailing_edge_angle"]

        # Load model options
        Cf = self.model_options["factors"]["skin_friction"]

        # Calculate exit velocity
        alpha_out = theta_out*0.9
        delta_M = np.exp(-Cf*(r_out-r_in)/(b_in*math.cosd(alpha_in)))
        # delta_M = 1
        v_t_out = v_t_in*r_in/r_out*delta_M
        v_out = v_t_out/math.sind(alpha_out)
        v_m_out = v_out*math.cosd(alpha_out)
        v_out = 0.2*a_in

        velocity_triangle_out = {
            "v" : v_out,
            "v_t" : v_t_out,
            "v_m" : v_m_out,
            "alpha" : alpha_out
        }
        
        # Calculate exit entropy
        d_out = mass_flow_rate/(v_out*math.cosd(alpha_out)*A_out)
        h0_out = h_in + 0.5*v_in**2
        h_out = h0_out - 0.5*v_out**2
        static_props_out = self.fluid.get_props(cp.DmassHmass_INPUTS, d_out, h_out)
        s_out = static_props_out["s"]

        initial_guess = {
                    "w_throat_vaned_diffuser" : input["v"]*0.5,
                    "v_out_vaned_diffuser" : v_out,
                    "s_out_vaned_diffuser" : s_out,
                    }
        
        exit_state = {
            **velocity_triangle_out,
            **static_props_out,
        }

        return initial_guess, exit_state

    def get_volute_guess(self, input):
        """
        Guess constant density
        Guess single velocity component at the exit
        
        """
        
        # Load input
        d_in = input["d"]
        h_in = input["h"]
        v_in = input["v"]

        # Load boundary conditions
        mass_flow_rate = self.boundary_conditions["mass_flow_rate"]

        # Load geometry
        A_out = self.geometry["volute"]["area_out"]

        # Calculate exit velocity
        d_out = d_in
        v_out = mass_flow_rate/(d_out*A_out)
        velocity_triangle_out = {
            "v" : v_out,
        }


        # Calculate exit entropy
        h0_out = h_in + 0.5*v_in**2
        h_out = h0_out - 0.5*v_out**2
        static_props_out = self.fluid.get_props(cp.DmassHmass_INPUTS, d_out, h_out)
        s_out = static_props_out["s"]

        initial_guess = {
                    "v_out_volute" : v_out,
                    "s_out_volute" : s_out,
                    }

        exit_plane = {
            **velocity_triangle_out,
            **static_props_out
        }

        return initial_guess, exit_plane


def print_simulation_summary(solvers):
    """
    Print a formatted footer summarizing the performance of all operation points.

    This function processes a list of solver objects to provide a summary of the performance
    analysis calculations. It calculates and displays the number of successful points and a summary of
    simulation tme statistics. Additionally, it lists the indices of failed operation points, if any.

    The function is robust against solvers that failed and lack certain attributes like 'elapsed_time'.
    In such cases, these solvers are included in the count of failed operation points, but not in the
    calculation time statistics.

    Parameters
    ----------
    solvers : list
        A list of solver objects. Each solver object should contain attributes related to the
        calculation of an operation point, such as 'elapsed_time' and the 'solution' status.

    """

    # Initialize times list and track failed points
    times = []
    failed_points = []

    for i, solver in enumerate(solvers):
        # Check if the solver is not None and has the required attribute
        if solver is not None and solver.success:
            times.append(solver.elapsed_time)
        else:
            # Handle failed solver or missing attributes
            failed_points.append(i)

    # Convert times to a numpy array for calculations
    times = np.asarray(times)
    total_points = len(solvers)

    # Define footer content
    width = 80
    separator = "-" * width
    lines_to_output = [
        "",
        separator,
        "Final summary of performance analysis calculations".center(width),
        separator,
        f" Simulation successful for {total_points - len(failed_points)} out of {total_points} points",
    ]

    # Add failed points message only if there are failed points
    if failed_points:
        lines_to_output.append(
            f"Failed operation points: {', '.join(map(str, failed_points))}"
        )

    # Add time statistics only if there are valid times
    if times.size > 0:
        lines_to_output.extend(
            [
                f" Average calculation time per operation point: {np.mean(times):.3f} seconds",
                f" Minimum calculation time of all operation points: {np.min(times):.3f} seconds",
                f" Maximum calculation time of all operation points: {np.max(times):.3f} seconds",
                f" Total calculation time for all operation points: {np.sum(times):.3f} seconds",
            ]
        )
    else:
        lines_to_output.append(" No valid calculation times available.")

    lines_to_output.append(separator)
    lines_to_output.append("")

    # Display to stdout
    for line in lines_to_output:
        print(line)


def print_boundary_conditions(BC):
    """
    Print the boundary conditions.

    This function prints the boundary conditions in a formatted manner. It takes a dictionary `BC`
    containing the following keys:

        - `fluid_name` (str): Name of the fluid.
        - `alpha_in` (float): Flow angle at inlet in degrees.
        - `T0_in` (float): Total temperature at inlet in Kelvin.
        - `p0_in` (float): Total pressure at inlet in Pascal.
        - `mass_flow_rate` (float): Static pressure at outlet in Pascal.
        - `omega` (float): Angular speed in radians per second.

    Parameters
    ----------
    BC : dict
        A dictionary containing the boundary conditions.

    """

    column_width = 25  # Adjust this to your desired width
    print("-" * 80)
    print(" Operating point: ")
    print("-" * 80)
    print(f" {'Fluid: ':<{column_width}} {BC['fluid_name']:<}")
    print(f" {'Flow angle in: ':<{column_width}} {BC['alpha_in']:<.2f} deg")
    print(f" {'Total temperature in: ':<{column_width}} {BC['T0_in']-273.15:<.2f} degC")
    print(f" {'Total pressure in: ':<{column_width}} {BC['p0_in']/1e5:<.3f} bar")
    print(f" {'Mass flow rate: ':<{column_width}} {BC['mass_flow_rate']:<.3f} kg/s")
    print(f" {'Angular speed: ':<{column_width}} {BC['omega']*60/2/np.pi:<.1f} RPM")
    print("-" * 80)
    print()


def print_operation_points(operation_points):
    """
    Prints a summary table of operation points scheduled for simulation.

    This function takes a list of operation point dictionaries, formats them
    according to predefined specifications, applies unit conversions where
    necessary, and prints them in a neatly aligned table with headers and units.

    Parameters
    ----------
    - operation_points (list of dict): A list where each dictionary contains
      key-value pairs representing operation parameters and their corresponding
      values.

    Notes
    -----
    - This function assumes that all necessary keys exist within each operation
      point dictionary.
    - The function directly prints the output; it does not return any value.
    - Unit conversions are hardcoded and specific to known parameters.
    - If the units of the parameters change or if different parameters are added,
      the unit conversion logic and `field_specs` need to be updated accordingly.
    """
    length = 80
    index_width = 8
    output_lines = [
        "-" * length,
        " Summary of operation points scheduled for simulation",
        "-" * length,
    ]

    # Configuration for each field with specified width and decimal places
    field_specs = {
        "fluid_name": {"name": "Fluid", "unit": "", "width": 8},
        "alpha_in": {"name": "angle_in", "unit": "[deg]", "width": 10, "decimals": 1},
        "T0_in": {"name": "T0_in", "unit": "[degC]", "width": 12, "decimals": 2},
        "p0_in": {"name": "p0_in", "unit": "[kPa]", "width": 12, "decimals": 2},
        "mass_flow_rate": {"name": "mass_flow_rate", "unit": "[kg/s]", "width": 12, "decimals": 2},
        "omega": {"name": "omega", "unit": "[RPM]", "width": 12, "decimals": 0},
    }

    # Create formatted header and unit strings using f-strings and field widths
    header_str = f"{'Index':>{index_width}}"  # Start with "Index" header
    unit_str = f"{'':>{index_width}}"  # Start with empty string for unit alignment

    for spec in field_specs.values():
        header_str += f" {spec['name']:>{spec['width']}}"
        unit_str += f" {spec['unit']:>{spec['width']}}"

    # Append formatted strings to the output lines
    output_lines.append(header_str)
    output_lines.append(unit_str)

    # Unit conversion functions
    def convert_units(key, value):
        if key == "T0_in":  # Convert Kelvin to Celsius
            return value - 273.15
        elif key == "omega":  # Convert rad/s to RPM
            return (value * 60) / (2 * np.pi)
        elif key == "alpha_in":  # Convert radians to degrees
            return np.degrees(value)
        elif key in ["p0_in"]:  # Pa to kPa
            return value / 1e3
        # elif key in ["mass_flow_rate"]:  # No change
        #     return value
        return value

    # Process and format each operation point
    for index, op_point in enumerate(operation_points, start=1):
        row = [f"{index:>{index_width}}"]
        for key, spec in field_specs.items():
            value = convert_units(key, op_point[key])
            if isinstance(value, float):
                # Format floats with the specified width and number of decimal places
                row.append(f"{value:>{spec['width']}.{spec['decimals']}f}")
            else:
                # Format strings to the specified width without decimals
                row.append(f"{value:>{spec['width']}}")
        output_lines.append(" ".join(row))  # Ensure spaces between columns

    output_lines.append("-" * length)  # Add a closing separator line

    # Join the lines and print the output
    formatted_output = "\n".join(output_lines)

    for line in output_lines:
        print(line)
    return formatted_output


def latin_hypercube_sampling(bounds, n_samples):
    """
    Generates samples using Latin Hypercube Sampling.

    Parameters:
    bounds (list of tuples): A list of (min, max) bounds for each variable.
    n_samples (int): The number of samples to generate.

    Returns:
    np.ndarray: An array of shape (n_samples, n_variables) containing the samples.
    """
    n_variables = len(bounds)
    # Create a Latin Hypercube Sampler
    sampler = qmc.LatinHypercube(d=n_variables, seed = 1)
    
    # Generate samples in the unit hypercube
    unit_hypercube_samples = sampler.random(n=n_samples)
    
    # Scale the samples to the provided bounds
    lower_bounds = np.array([b[0] for b in bounds])
    upper_bounds = np.array([b[1] for b in bounds])
    scaled_samples = qmc.scale(unit_hypercube_samples, lower_bounds, upper_bounds)

    return scaled_samples




from pydantic import (
    BaseModel,
    ValidationError,
    Field,
    model_validator,
    field_validator,
    ConfigDict,
)
from typing import Optional, List, Union, Literal, Tuple, Dict
from enum import Enum
from typing_extensions import Self
from typing_extensions import Annotated
from pydantic.functional_validators import AfterValidator, BeforeValidator
from ..axial_turbine import (
    LOSS_MODELS,
    LOSS_COEFFICIENTS,
    CHOKING_CRITERIONS,
    DEVIATION_MODELS,
    RADIUS_TYPE,
)
from ..pysolver_view import (
    SOLVER_OPTIONS,
    VALID_LIBRARIES_AND_METHODS,
    DERIVATIVE_METHODS,
)
import re


class OperationPoints(BaseModel):
    """
    Model describing the operational points of an axial turbine.

    Attributes
    ----------
    fluid_name : str
        Name of the fluid.
    T0_in : float
        Inlet total temperature (must be greater than 0).
    p0_in : float
        Inlet total pressure (must be greater than 0).
    p_out : float
        Outlet pressure (must be greater than 0).
    omega : float
        Rotational speed (must be greater than 0).
    alpha_in : float
        Inlet angle.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid")
    fluid_name: str
    T0_in: Annotated[float, Field(gt=0)]
    p0_in: Annotated[float, Field(gt=0)]
    p_out: Annotated[float, Field(gt=0)]
    omega: Annotated[float, Field(gt=0)]
    alpha_in: float


class PerformanceMap(BaseModel):
    """
    Model describing a performance map for axial turbine analysis.

    Attributes
    ----------
    fluid_name : Union[str, List[str]]
        Name of the fluid or list of fluid names.
    T0_in : float, list
        Inlet total temperature or list of inlet total temperatures (each must be greater than 0).
    p0_in : float, list
        Inlet total pressure or list of inlet total pressures (each must be greater than 0).
    p_out : float, list
        Outlet pressure or list of outlet pressures (each must be greater than 0).
    omega : float, list
        Rotational speed or list of rotational speeds (each must be greater than 0).
    alpha_in : float, list
        Inlet angle or list of inlet angles.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid")
    fluid_name: Union[str, List[str]]
    T0_in: Union[Annotated[float, Field(gt=0)], List[Annotated[float, Field(gt=0)]]]
    p0_in: Union[Annotated[float, Field(gt=0)], List[Annotated[float, Field(gt=0)]]]
    p_out: Union[Annotated[float, Field(gt=0)], List[Annotated[float, Field(gt=0)]]]
    omega: Union[Annotated[float, Field(gt=0)], List[Annotated[float, Field(gt=0)]]]
    alpha_in: Union[float, List[float]]


class TuningFactors(BaseModel):
    """
    Model describing tuning factors of loss coefficients.

    Attributes
    ----------
    profile : float, optional
        Tuning factor for profile loss coefficient. Default is 1.00.
    incidence : float, optional
        Tuning factor for incidence loss coefficient. Default is 1.00.
    secondary : float, optional
        Tuning factor for secondary loss coefficient. Default is 1.00.
    trailing : float, optional
        Tuning factor for trailing loss coefficient. Default is 1.00.
    clearance : float, optional
        Tuning factor for clearance loss coefficient. Default is 1.00.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid")
    profile: float = 1.00
    incidence: float = 1.00
    secondary: float = 1.00
    trailing: float = 1.00
    clearance: float = 1.00

LossModelEnum = Enum('LossModels', dict(zip([model.upper() for model in LOSS_MODELS], LOSS_MODELS)))
LossCoefficientEnum = Enum('LossCoefficients', dict(zip([model.upper() for model in LOSS_COEFFICIENTS], LOSS_COEFFICIENTS)))

class LossModel(BaseModel):
    """
    Model describing loss model setup and related parameters.

    Attributes
    ----------
    model : str
        The type of loss model. Should be one of the predefined loss models.
    loss_coefficient : str
        The loss coefficient type. Should be one of the predefined loss coefficients.
    custom_value : float
        Loss coefficient value if model is 'custom'
    inlet_displacement_thickness_height_ratio : float, optional
        Ratio of inlet displacement thickness to height. Default is 0.011.
    tuning_factors : TuningFactors, optional
        Object containing tuning factors for loss coefficients. Default is a new instance of TuningFactors.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid", use_enum_values=True)
    model : LossModelEnum
    loss_coefficient : LossCoefficientEnum
    custom_value: Annotated[float, Field(ge=0)] = 0.0
    inlet_displacement_thickness_height_ratio: float = 0.011
    tuning_factors: TuningFactors = TuningFactors()

DeviationModelEnum = Enum('DeviationModels', dict(zip([model.upper() for model in DEVIATION_MODELS], DEVIATION_MODELS)))
ChokingCriterionEnum = Enum('ChokingModels', dict(zip([model.upper() for model in CHOKING_CRITERIONS], CHOKING_CRITERIONS)))

class SimulationOptions(BaseModel):
    """
    Model describing simulation options.

    Attributes
    ----------
    deviation_model : str
        The type of deviation model. Should be one of the predefined deviation models.
    blockage_model : str, float, optional
        The blockage model. Can be a string or a float value. Default is 0.00.
    choking_model : str
        The choking model. Should be one of the predefined choking models.
    rel_step_fd : float, optional
        Relative step for finite difference calculations (only for evaluate_cascade_critical choking model). Default is 1e-4.
    loss_model : LossModel
        Object describing loss model setup and related parameters.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid", use_enum_values=True)
    deviation_model : DeviationModelEnum
    blockage_model: Union[str, float] = 0.00
    choking_criterion : ChokingCriterionEnum
    rel_step_fd: float = 1e-4
    loss_model: LossModel

SolverOptionEnum = Enum('SolverOptions', dict(zip([model.upper() for model in SOLVER_OPTIONS], SOLVER_OPTIONS)))
DerivativeMethodEnum = Enum('DerivativeMethods', dict(zip([model.upper() for model in DERIVATIVE_METHODS], DERIVATIVE_METHODS)))

class SolverOptionsPerformanceAnalysis(BaseModel):
    """
    Model describing solver options for performance analysis.

    Attributes
    ----------
    method : str, optional
        The root solver method. Should be one of the predefined solver options. Default is 'hybr'.
    tolerance : float, optional
        Tolerance for convergence. Default is 1e-8.
    max_iterations : int, optional
        Maximum number of iterations. Default is 100.
    derivative_method : str, optional
        The method for finite difference approximation calculation. Should be one of the predefined derivative methods. Default is '2-point'.
    derivative_abs_step : float, optional
        Absolute step size for finite difference approximation calculation. Default is 1e-6.
    print_convergence : bool, optional
        Flag indicating whether to print convergence information. Default is True.
    plot_convergence : bool, optional
        Flag indicating whether to plot convergence. Default is False.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid", use_enum_values=True)
    method : SolverOptionEnum = SOLVER_OPTIONS[0]
    tolerance: float = 1e-8
    max_iterations: int = 100
    derivative_method : DerivativeMethodEnum = DERIVATIVE_METHODS[0]
    derivative_abs_step: float = 1e-6
    print_convergence: bool = True
    plot_convergence: bool = False

CascadeTypeEnum = Enum('CascadeTypes', dict(zip([model.upper() for model in ["stator", "rotor"]], ["stator", "rotor"])))

class GeometryPerformanceAnalysis(BaseModel):
    """
    Model describing geometry parameters for performance analysis.

    Attributes
    ----------
    cascade_type : list[str]
        Types of cascades, either 'stator' or 'rotor'.
    radius_hub_in : list[float]
        Hub radius of the cascades inlet (each must be greater than 0).
    radius_hub_out : list[float]
        Hub radius of the cascades exit (each must be greater than 0).
    radius_tip_in : list[float]
        Tip radius of the cascades inlet (each must be greater than 0).
    radius_tip_out : list[float]
        Tip radius of the cascades exit (each must be greater than 0).
    pitch : list[float]
        Pitch of the cascades (each must be greater than 0).
    chord : list[float]
        Chord of the cascades (each must be greater than 0).
    stagger_angle : list[float]
        Stagger angle of the cascades, in degrees (each must be between -90 and 90).
    opening : list[float]
        Throat opening of the cascades (each must be greater than 0).
    leading_edge_angle : list[float]
        Leading edge angle of the cascades, in degrees (each must be between -90 and 90).
    leading_edge_wedge_angle : list[float]
        Leading edge wedge angle of the cascades, in degrees (each must be between 0 and 90).
    leading_edge_diameter : list[float]
        Diameter of the leading edge of the cascades (each must be greater than 0).
    trailing_edge_thickness : list[float]
        Thickness of the trailing edge of the cascades (each must be greater than 0).
    maximum_thickness : list[float]
        Maximum thickness of the cascades (each must be greater than 0).
    tip_clearance : list[float]
        Tip clearance of the cascades (each must be greater than or equal 0).
    throat_location_fraction : list[float]
        Fractional location of the throat in the cascades (each must be between 0 and 1).

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid", use_enum_values=True)
    cascade_type: List[CascadeTypeEnum]
    radius_hub_in: List[Annotated[float, Field(gt=0)]]
    radius_hub_out: List[Annotated[float, Field(gt=0)]]
    radius_tip_in: List[Annotated[float, Field(gt=0)]]
    radius_tip_out: List[Annotated[float, Field(gt=0)]]
    pitch: List[Annotated[float, Field(gt=0)]]
    chord: List[Annotated[float, Field(gt=0)]]
    stagger_angle: List[Annotated[float, Field(le=90), Field(gt=-90)]]
    opening: List[Annotated[float, Field(gt=0)]]
    leading_edge_angle: List[Annotated[float, Field(le=90), Field(gt=-90)]]
    leading_edge_wedge_angle: List[Annotated[float, Field(le=90), Field(gt=0)]]
    leading_edge_diameter: List[Annotated[float, Field(gt=0)]]
    trailing_edge_thickness: List[Annotated[float, Field(gt=0)]]
    maximum_thickness: List[Annotated[float, Field(gt=0)]]
    tip_clearance: List[Annotated[float, Field(ge=0)]]
    throat_location_fraction: List[Annotated[float, Field(le=1), Field(gt=0)]]

    @model_validator(mode="after")
    def check_length(self) -> Self:
        attributes = vars(self).values()
        if not all(len(attr) == len(self.cascade_type) for attr in attributes):
            raise ValueError("Geometry input is not of same length")
        return self

LibraryEnum = Enum('LibraryOptions', dict(zip([model.upper() for model in VALID_LIBRARIES_AND_METHODS.keys()], VALID_LIBRARIES_AND_METHODS.keys())))
MethodEnum = Enum('MethodOption', dict(zip([model.upper() for model in [x for xs in VALID_LIBRARIES_AND_METHODS.values() for x in xs]], [x for xs in VALID_LIBRARIES_AND_METHODS.values() for x in xs])))
UpdateOnEnum = Enum('UpdateOns', dict(zip([model.upper() for model in ["gradient", "function"]], ["gradient", "function"])))

class SolverOptionsOptimization(BaseModel):
    """
    Model describing solver options for optimization.

    Checks that the given solver method corresponds with the given library.

    Attributes
    ----------
    library : str, optional
        The optimization library to be used. Should be one of the predefined libraries. Default is 'scipy'.
    method : str, optional
        The optimization method to be used. Default is 'slsqp'.
    tolerance : float, optional
        Tolerance for convergence. Default is 1e-5.
    max_iterations : int, optional
        Maximum number of iterations. Default is 100.
    derivative_method : str, optional
        The method for finite difference calculation. Should be one of the predefined derivative methods. Default is '2-point'.
    derivative_abs_step : float, optional
        Absolute step size for finite difference calculation. Default is None.
    print_convergence : bool, optional
        Flag indicating whether to print convergence information. Default is True.
    plot_convergence : bool, optional
        Flag indicating whether to plot convergence. Default is False.
    update_on : str, optional
        The criteria for updating. Should be either 'gradient' or 'function'. Default is 'gradient'.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid", use_enum_values=True)
    library: LibraryEnum = list(VALID_LIBRARIES_AND_METHODS.keys())[0]
    method: MethodEnum = VALID_LIBRARIES_AND_METHODS[list(VALID_LIBRARIES_AND_METHODS.keys())[0]][0]
    derivative_method: DerivativeMethodEnum = DERIVATIVE_METHODS[0]
    derivative_abs_step: float = None
    print_convergence: bool = True
    plot_convergence: bool = False
    update_on: UpdateOnEnum = "gradient"
    options : Dict = {}

    @model_validator(mode="after")
    def check_solver_method(self) -> Self:
        lib = self.library
        method = self.method
        if not method in VALID_LIBRARIES_AND_METHODS[lib]:
            raise ValueError(f"Method {method} is not available in {lib} library.")
        return self

class InitialGuess2(BaseModel):

    model_config = ConfigDict(extra="allow")
    efficiency_tt : Union[float, List[float]] = [0.9, 0.8]
    efficiency_ke : Union[float, List[float]] = [0.8, 0.7]

    @model_validator(mode="after")
    def check_length(self) -> Self:
        attributes = vars(self).values()
        input_type = type(self.efficiency_tt)
        print(attributes)
        if not all(isinstance(attr, input_type) for attr in attributes):
            raise ValueError("Variable input is not of same type")
        if input_type == list:
            if not all(len(attr) == len(self.efficiency_tt) for attr in attributes):
                raise ValueError("Variable input is not of same length")
        return self
    
class PerformanceAnalysis(BaseModel):
    """
    Model describing performance analysis parameters.

    Attributes
    ----------
    performance_map : PerformanceMap, optional
        Object containing performance map data.
    solver_options : SolverOptionsPerformanceAnalysis, optional
        Object containing solver options for performance analysis.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid")
    performance_map: PerformanceMap = None
    solver_options: SolverOptionsPerformanceAnalysis = (
        SolverOptionsPerformanceAnalysis()
    )
    initial_guess : Dict = {}

    @model_validator(mode='before')
    def validate_initial_guess_keys(cls, v):
        allowed_keys = {'efficiency_tt', 'efficiency_ke', 'n_samples', 'ma'}
        ma_pattern = re.compile(r'^ma_(\d+)$')
        ma_indices = set()
        
        initial_guess = v.get('initial_guess', {})
        for key in initial_guess.keys():
            if key not in allowed_keys:
                match = ma_pattern.match(key)
                if not match:
                    raise ValueError(f'Invalid key in initial_guess: {key}. Allowed keys are efficiency_tt, efficiency_ke, and ma_<number>')
                ma_index = int(match.group(1))
                if ma_index in ma_indices:
                    raise ValueError(f'Duplicate ma_i index {ma_index} in initial_guess')
                ma_indices.add(ma_index)
        
        # Check if ma_indices include all numbers from 1 to len(ma_indices)
        if ma_indices and (min(ma_indices) != 1 or max(ma_indices) != len(ma_indices)):
            raise ValueError(f'ma_i indices must include all numbers from 1 to {len(ma_indices)} without gaps')
        return v


class Variable(BaseModel):
    """
    Model describing a variable.

    If the attributes are lists, each value corresponds to a seperate cascade.

    Checks that the type and length of the attributes are the same.

    Attributes
    ----------
    value : float, str, list[float], list[str]
        The value(s) of the variable, which could be a float, string, list of floats, or list of strings.
    lower_bound : float, list[float], optional
        The lower bound of the variable. Default is None.
    upper_bound : float, list[float], optional
        The upper bound of the variable. Default is None.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid", use_enum_values=True)
    value : Union[float, CascadeTypeEnum, List[float], List[CascadeTypeEnum]]
    lower_bound: Union[float, List[float]] = None
    upper_bound: Union[float, List[float]] = None

    @model_validator(mode="after")
    def check_length(self) -> Self:
        attributes = vars(self).values()
        input_type = type(self.value)
        if self.lower_bound is not None:
            if not all(isinstance(attr, input_type) for attr in attributes):
                raise ValueError("Variable input is not of same type")
            if isinstance(self.value, list):
                if not all(len(attr) == len(self.value) for attr in attributes):
                    raise ValueError("Variable input is not of same length")
        return self


class Constraint(BaseModel):
    """
    Model describing a constraint.

    Attributes
    ----------
    type : str
        The type of constraint, indicating whether it is less than ('<'), equal to ('='), or greater than ('>').
    value : float
        The value of the constraint.
    normalize : bool
        Flag indicating whether the constraint should be normalized.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid")
    variable : str
    type: Literal["<", "=", ">"]
    value: Union[float, List[float]]
    normalize: Union[bool, float]

RadiusTypeEnum = Enum('RadiusTypes', dict(zip([model.upper() for model in RADIUS_TYPE], RADIUS_TYPE)))

class ObjectiveFunction(BaseModel):
    """
    Model describing the objective function of a optimization problem.

    Attributes
    ----------
    variable : str
        Defines the objective function variable. Must be contained in the `results` structure of the problem and be on the form 'section.parameter', e.g. 'overall.efficiency_ts'.
    type : str
        Whether the objective function should be maxmized or minimized.
    scale : float
        Value to scale the objective function.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid")
    variable : str = 'overall.efficieny_ts'
    type: Literal["maximize", "minimize"] = 'maximize'
    scale: float = 1

class Variables(BaseModel):

    model_config = ConfigDict(extra="forbid")
    v_in : Variable = Variable(value = 0.1, lower_bound = 0.001, upper_bound=0.5)
    w_out : Variable = Variable(value = [0.65, 0.65], lower_bound = [0.1, 0.1], upper_bound=[1.0, 1.0])
    s_out : Variable = Variable(value = [0.15, 0.15], lower_bound = [0.0, 0.0], upper_bound=[0.32, 0.32])
    beta_out : Variable = Variable(value = [0.83, 0.17], lower_bound = [0.72, 0.06], upper_bound=[0.94, 0.28])
    v_crit_in : Variable = Variable(value = [0.4, 0.4], lower_bound = [0.1, 0.1], upper_bound=[1.0, 1.0])
    beta_crit_throat : Variable = Variable(value = [0.83, 0.17], lower_bound = [0.72, 0.06], upper_bound=[0.94, 0.28])
    w_crit_throat : Variable = Variable(value = [0.65, 0.65], lower_bound = [0.1, 0.1], upper_bound=[1.0, 1.0])
    s_crit_throat : Variable = Variable(value = [0.15, 0.15], lower_bound = [0.0, 0.0], upper_bound=[0.32, 0.32])
    specific_speed : Variable = Variable(value = 1.2, lower_bound=0.01, upper_bound=10)
    blade_jet_ratio : Variable = Variable(value = 0.5)
    hub_tip_ratio_in : Variable = Variable(value = [0.6, 0.6])
    hub_tip_ratio_out : Variable = Variable(value = [0.6, 0.6])
    aspect_ratio : Variable = Variable(value = [1.5, 1.5])
    pitch_chord_ratio : Variable = Variable(value = [0.9, 0.9])
    trailing_edge_thickness_opening_ratio : Variable = Variable(value = [0.1, 0.1])
    leading_edge_angle : Variable = Variable(value = [0.41, 0.5])
    gauging_angle : Variable = Variable(value = [0.17, 0.94])
    throat_location_fraction : Variable = Variable(value = [1.0, 1.0])
    leading_edge_diameter : Variable = Variable(value = [2*0.00600448, 2*0.00600448])
    leading_edge_wedge_angle : Variable = Variable(value = [45.00, 45.00] )
    tip_clearance : Variable = Variable(value = [0.00, 5e-4])
    cascade_type : Variable = Variable(value = ["stator", "rotor"])

class DesignOptimization(BaseModel):
    """
    Model describing design optimization parameters.

    Checks that variables contain all variables from a predifned list (VARIABLES). Returns a default dictionary if not.

    Attributes
    ----------
    objective_function : str, optional
        The objective function for optimization. Should be one of the predefined objective functions. Default is 'efficiency_ts'.
    variables : dict[str], optional
        Dictionary containing variables for optimization. Should contain a dict for each variable in a predefined variable list (VARIABLES).
    constraints : dict[str], optional
        Dictionary containing constraints for optimization.
    solver_options : SolverOptionsOptimization, optional
        Object containing solver options for optimization.
    radius_type : str, optional
        The type of radius used in optimization. Should be one of the predefined radius types. Default is 'constant_mean'.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid", use_enum_values=True)
    objective_function : ObjectiveFunction = ObjectiveFunction()
    variables : Variables = Variables()
    constraints : List[Constraint] = None
    solver_options: SolverOptionsOptimization = SolverOptionsOptimization()
    radius_type : RadiusTypeEnum = RADIUS_TYPE[0]
    multistarts : int = 0 



class AxialTurbine(BaseModel):
    """
    Model describing an axial turbine.

    Attributes
    ----------
    turbomachinery : str
        The type of turbomachinery, which for this class is "axial_turbine".
    operation_points : OperationPoints
        Object containing operation points data.
    simulation_options : SimulationOptions
        Object containing simulation options data.
    geometry : GeometryPerformanceAnalysis, optional
        Object containing geometry performance analysis data. Default is None.
    performance_analysis : PerformanceAnalysis, optional
        Object containing performance analysis data. Default is a new instance of PerformanceAnalysis().
    design_optimization : DesignOptimization, optional
        Object containing design optimization data. Default is a new instance of DesignOptimization().

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid")
    turbomachinery: Literal["axial_turbine"]
    operation_points: Union[List[OperationPoints], OperationPoints]
    simulation_options: SimulationOptions
    geometry: GeometryPerformanceAnalysis = None
    performance_analysis: PerformanceAnalysis = PerformanceAnalysis()
    design_optimization: DesignOptimization = DesignOptimization()




from .. import utilities as utils
from . import loss_model_custom as lmc
from . import loss_model_oh as lmo
from . import loss_model_zhang_set1 as zs1
from . import loss_model_zhang_set2 as zs2
from . import loss_model_zhang_set3 as zs3


# List of valid options
LOSS_MODELS = [
    "isentropic",
    "custom",
    "oh",
    "zhang_set1",
    "zhang_set2",
    "zhang_set3",
]
"""
Available loss models.
"""

LOSS_COEFFICIENTS = [
    "stagnation_enthalpy_loss",
    "static_enthalpy_loss"
]
"""
Available loss coefficients.
"""

# Keys that the output loss dictionary must have
KEYS_LOSSES = [
]


def evaluate_loss_model(loss_model_options, input_parameters, component):
    """
    Calculate loss coefficient based on the selected loss model.

    The avaialble loss models are:

    - `isentropic` : Loss coefficient set to zero (isentropic).
    - `custom` :  Constant loss coefficent according to user. Require `loss_model_options["custom_value"]`.

    Parameters
    ----------
    loss_model_options : dict
        Options for the loss calculation.
    input_parameters : dict
        Input parameters required for loss model calculation.

    Returns:
    dict
        A dictionary containing loss components.
    """

    # Function mappings for each loss model
    loss_model_functions = {
        LOSS_MODELS[0]: lambda x,y: {
            "loss_total": 0.0,
        },
        LOSS_MODELS[1] : lmc.compute_losses,
        LOSS_MODELS[2] : lmo.compute_losses,
        LOSS_MODELS[3] : zs1.compute_losses,
        LOSS_MODELS[4] : zs2.compute_losses,
        LOSS_MODELS[5] : zs3.compute_losses,

    }

    # Evaluate loss model
    model = loss_model_options["model"]
    if model in loss_model_functions:
        loss_dict = loss_model_functions[model](input_parameters, component)
    else:
        options = ", ".join(f"'{k}'" for k in LOSS_MODELS)
        raise ValueError(f"Invalid loss model '{model}'. Available options: {options}")

    # Apply tuning factors (empty dict if not provided)
    tuning_factors = loss_model_options.get("tuning_factors", {})
    tuning_factors = {f"loss_{key}": value for key, value in tuning_factors.items()}
    apply_tuning_factors(loss_dict, tuning_factors)

    # Compute the loss coefficient according to definition
    loss_coeff = loss_model_options["loss_coefficient"]
    if loss_coeff == "stagnation_enthalpy_loss":
        h0rel_out_is = input_parameters["h0_rel_out_is"]
        h0rel_out = input_parameters["h0_rel_out"]
        Y_definition = h0rel_out - h0rel_out_is
    elif loss_coeff == "static_enthalpy_loss":
        # h_out_is = input_parameters["h_out_is"]
        # h_out = input_parameters["h_out"]
        h_out_is = input_parameters["isentropic"]["h"]
        h_out = input_parameters["exit_plane"]["h"]
        v_in = input_parameters["inlet_plane"]["v"]
        Y_definition = 2*(h_out - h_out_is)/(v_in**2)

    else:
        options = ", ".join(f"'{k}'" for k in LOSS_COEFFICIENTS)
        raise ValueError(
            f"Invalid loss coefficient '{loss_coeff}'. Available options: {options}"
        )

    # Compute loss coefficient error
    loss_dict["loss_error"] = Y_definition - loss_dict["loss_total"]

    return loss_dict


def apply_tuning_factors(loss_dict, tuning_factors):
    """
    Apply tuning factors to the loss model

    The tuning factors are multiplied with their associated loss component.

    Parameters
    ----------
    loss_dict : dict
        A dictionary containing loss components.
    tuning_factors : dict
        A dictionary containing the multiplicative tuning factors.

    Raises:
        KeyError: If a key from `tuning_factors` is not found in `loss_dict`.
    """

    for key, factor in tuning_factors.items():
        if key not in loss_dict:
            raise KeyError(f"Tuning factor key '{key}' not found in loss dictionary.")
        loss_dict[key] *= factor

    return loss_dict